"""Plugin facility to read and normalize information from different sources to a common format."""

import warnings
import re

import numpy as np
import xarray as xr
import h5py
from astropy.io import fits

from pathlib import Path
from typing import Any
import copy
import os.path

import arpes.config
from arpes.utilities.dict import case_insensitive_get
from arpes.utilities.xarray import rename_dataset_keys
from .utilities.hdf5 import (
    construct_coords,
    get_attrs,
    dataset_to_array,
)
from .utilities.fits import find_clean_coords

__all__ = [
    "endstation_name_from_alias",
    "endstation_from_alias",
    "add_endstation",
    "load_scan",
    "EndstationBase",
    "HDF5Endstation",
    "FITSEndstation",
    "HemisphericalEndstation",
    "SynchrotronEndstation",
    "SingleFileEndstation",
    "resolve_endstation",
]

_ENDSTATION_ALIASES = {}


class EndstationBase:
    """Implements the core features of ARPES data loading.

    A thorough documentation
    is available at `the plugin documentation <https://arpes.readthedocs.io/writing-plugins>`_.

    To summarize, a plugin has a few core jobs:

    1. Load data, including collating any data that is in a multi-file format
       This is accomplished with `.load`, which delegates loading `frames` (single files)
       to `load_single_frame`. Frame collation is then performed by `concatenate_frames`.
    2. Loading and attaching metadata.
    3. Normalizing metadata to standardized names. These are documented at the
       `data model documentation <https://arpes.readthedocs.io/spectra>`_.
    4. Ensuring all angles and necessary coordinates are attached to the data.
       Data should permit immediate conversion to angle space after being loaded.

    Plugins are in one-to-many correspondance with the values of the "location" column in
    analysis spreadsheets. This binding is provided by PRINCIPAL_NAME and ALIASES.

    The simplest way to normalize metadata is by renaming keys, but sometimes additional
    work is required. RENAME_KEYS is provided to make this simpler, and is implemented in
    scan post-processessing.
    """

    ALIASES = []
    PRINCIPAL_NAME = None
    ATTR_TRANSFORMS = {}
    MERGE_ATTRS = {}

    _SEARCH_DIRECTORIES = (
        "",
        "hdf5",
        "fits",
        "../Data",
        "../Data/hdf5",
        "../Data/fits",
    )
    _SEARCH_PATTERNS = (
        r"[\-a-zA-Z0-9_\w]+_[0]+{}$",
        r"[\-a-zA-Z0-9_\w]+_{}$",
        r"[\-a-zA-Z0-9_\w]+{}$",
        r"[\-a-zA-Z0-9_\w]+[0]{}$",
    )
    _TOLERATED_EXTENSIONS = {
        ".h5",
        ".nc",
        ".fits",
        ".pxt",
        ".nxs",
        ".txt",
    }
    _USE_REGEX = True

    # adjust as needed
    ENSURE_COORDS_EXIST = ["x", "y", "z", "theta", "beta", "chi", "hv", "alpha", "psi"]
    CONCAT_COORDS = ["hv", "chi", "psi", "timed_power", "tilt", "beta", "theta"]

    # phi because this happens sometimes at BL4 with core level scans
    SUMMABLE_NULL_DIMS = ["phi", "cycle"]

    RENAME_KEYS = {}

    @classmethod
    def is_file_accepted(cls, file, scan_desc) -> bool:
        """Determines whether this loader can load this file."""
        if os.path.exists(str(file)) and len(str(file).split(os.path.sep)) > 1:
            # looks like an actual file, we are going to just check that the extension is kosher
            # and that the filename matches something reasonable.
            p = Path(str(file))

            if p.suffix not in cls._TOLERATED_EXTENSIONS:
                return False

            for pattern in cls._SEARCH_PATTERNS:
                regex = re.compile(pattern.format(r"[0-9]+"))
                if regex.match(p.stem):
                    return True

            return False
        try:
            _ = cls.find_first_file(file, scan_desc)
            return True
        except ValueError:
            return False

    @classmethod
    def files_for_search(cls, directory) -> list[str]:
        """Filters files in a directory for candidate scans.

        Here, this just means collecting the ones with extensions acceptable to the loader.
        """
        return [
            f
            for f in os.listdir(directory)
            if os.path.splitext(f)[1] in cls._TOLERATED_EXTENSIONS
        ]

    @classmethod
    def find_first_file(cls, file, scan_desc, allow_soft_match=False):
        """Attempts to find a file associated to the scan given the user provided path or scan number.

        This is mostly done by regex matching over available options.
        Endstations which do not require further control here can just provide class attributes:

        * `._SEARCH_DIRECTORIES`: Defining which paths should be checked for scans
        * `._SEARCH_PATTERNS`: Defining acceptable filenames
        * `._USE_REGEX`: Controlling literal or regex filename checking
        * `._TOLERATED_EXTENSIONS`: Controlling whether files should be rejected based on their extension.
        """
        base_dir = getattr(arpes.config, "data_root", os.getcwd())
        dir_options = [
            os.path.join(base_dir, option) for option in cls._SEARCH_DIRECTORIES
        ]

        # another plugin related option here is we can restrict the number of regexes by allowing plugins
        # to install regexes for particular endstations, if this is needed in the future it might be a good way
        # of preventing clashes where there is ambiguity in file naming scheme across endstations

        patterns = [re.compile(m.format(file)) for m in cls._SEARCH_PATTERNS]

        for dir in dir_options:
            try:
                files = cls.files_for_search(dir)

                if cls._USE_REGEX:
                    for p in patterns:
                        for f in files:
                            m = p.match(os.path.splitext(f)[0])
                            if m is not None:
                                if m.string == os.path.splitext(f)[0]:
                                    return os.path.join(dir, f)
                else:
                    for f in files:
                        if os.path.splitext(file)[0] == os.path.splitext(f)[0]:
                            return os.path.join(dir, f)
                        if allow_soft_match:
                            matcher = os.path.splitext(f)[0].split("_")[-1]
                            try:
                                if int(matcher) == int(file):
                                    return os.path.join(dir, f)  # soft match
                            except ValueError:
                                pass
            except FileNotFoundError:
                pass

        if str(file) and str(file)[0] == "f":  # try trimming the f off
            return cls.find_first_file(
                str(file)[1:], scan_desc, allow_soft_match=allow_soft_match
            )

        raise ValueError("Could not find file associated to {}".format(file))

    def concatenate_frames(self, frames=list[xr.Dataset], scan_desc: dict = None):
        """Performs concatenation of frames in multi-frame scans.

        The way this happens is that we look for an axis on which the frames are changing uniformly
        among a set of candidates (`.CONCAT_COORDS`). Then we delegate to xarray to perform the concatenation
        and clean up the merged coordinate.
        """
        if not frames:
            raise ValueError("Could not read any frames.")

        if len(frames) == 1:
            return frames[0]

        # determine which axis to stitch them together along, and then do this
        scan_coord = None
        max_different_values = -np.inf
        for possible_scan_coord in self.CONCAT_COORDS:
            coordinates = [f.attrs.get(possible_scan_coord, None) for f in frames]
            n_different_values = len(set(coordinates))
            if n_different_values > max_different_values and None not in coordinates:
                max_different_values = n_different_values
                scan_coord = possible_scan_coord

        assert scan_coord is not None

        for frame in frames:
            frame.coords[scan_coord] = frame.attrs[scan_coord]

        frames.sort(key=lambda x: x.coords[scan_coord])
        return xr.concat(frames, scan_coord)

    def resolve_frame_locations(self, scan_desc: dict = None) -> list[str]:
        """Determine all files and frames associated to this piece of data.

        This always needs to be overridden in subclasses to handle data appropriately.
        """
        raise NotImplementedError(
            "You need to define resolve_frame_locations or subclass SingleFileEndstation."
        )

    def load_single_frame(
        self, frame_path: str = None, scan_desc: dict = None, **kwargs
    ) -> xr.Dataset:
        """Hook for loading a single frame of data.

        This always needs to be overridden in subclasses to handle data appropriately.
        """
        return xr.Dataset()

    def postprocess_frame(self, frame: xr.Dataset):
        """Performs frame level normalization of scan data.

        Here, we currently:
        1. Remove dimensions if they only have a single point, i.e. if the scan has shape [1,N] it
          gets squeezed to have size [N]
        2. Rename attributes
        """

        frame = rename_dataset_keys(frame, self.RENAME_KEYS)

        sum_dims = []
        for dim in frame.dims:
            if len(frame.coords[dim]) == 1 and dim in self.SUMMABLE_NULL_DIMS:
                sum_dims.append(dim)

        if sum_dims:
            frame = frame.sum(sum_dims, keep_attrs=True)

        return frame

    def postprocess_scan(self, data: xr.Dataset, scan_desc: dict = None):
        """Perform final normalization of scan data.

        This defines the common codepaths for attaching extra information to scans at load time.
        Currently this means we:

        1. Attach a normalized "type" or "kind" of the spectrum indicating what sort of scan it is
        2. Ensure standard coordinates are represented
        3. Apply attribute renaming and attribute transformations defined by class attrs
        4. Ensure the scan endianness matches the system for performance reasons down the line
        """

        for k, key_fn in self.ATTR_TRANSFORMS.items():
            if k in data.attrs:
                transformed = key_fn(data.attrs[k])
                if isinstance(transformed, dict):
                    data.attrs.update(transformed)
                else:
                    data.attrs[k] = transformed
        for k, v in self.MERGE_ATTRS.items():
            if k not in data.attrs:
                data.attrs[k] = v
        for coord in self.ENSURE_COORDS_EXIST:
            if coord not in data.coords:
                if coord in data.data_vars:
                    value = data.data_vars[coord]
                    data.drop_vars(coord)
                    data.coords[coord] = value
                elif coord in data.attrs:
                    data.coords[coord] = data.attrs[coord]
                else:
                    warnings.warn(
                        f"Could not assign coordinate {coord} from attributes, assigning np.nan instead."
                    )
                    data.coords[coord] = np.nan
        if "chi" in data.coords and "chi_offset" not in data.attrs:
            data.attrs["chi_offset"] = data.coords["chi"].item()

        # go and change endianness and datatypes to something reasonable
        # this is done for performance reasons in momentum space conversion, primarily
        for data_var in data.data_vars.values():
            if not data_var.dtype.isnative:
                data_var.values = data_var.values.view(
                    data_var.values.dtype.newbyteorder("=")
                )
                data_var.values = data_var.values.byteswap()

        return data

    def load_from_path(self, path: str | Path) -> xr.Dataset:
        """Convenience wrapper around `.load` which references an explicit path."""
        path = str(path)
        return self.load(
            {
                "file": path,
                "location": self.PRINCIPAL_NAME,
            }
        )

    def load(self, scan_desc: dict = None, **kwargs) -> xr.Dataset:
        """Loads a scan from a single file or a sequence of files.

        This defines the contract and structure for standard data loading plugins:
        1. Search for files (`.resolve_frame_locations`)
        2. Load them sequentially (`.load_single_frame`)
        3. Apply cleaning code to each frame (`.postprocess`)
        4. Concatenate these loaded files  (`.concatenate_frames`)
        5. Apply postprocessing code to the concatenated dataset

        You can read more about the plugin system in the detailed documentation,
        but for the most part loaders just specializing one or more of these different steps
        as appropriate for a beamline.
        """
        resolved_frame_locations = self.resolve_frame_locations(scan_desc)
        resolved_frame_locations = [
            f if isinstance(f, str) else str(f) for f in resolved_frame_locations
        ]
        frames = [
            self.load_single_frame(fpath, scan_desc, **kwargs)
            for fpath in resolved_frame_locations
        ]
        frames = [self.postprocess_frame(f) for f in frames]
        concatted = self.concatenate_frames(frames, scan_desc)
        concatted = self.postprocess_scan(concatted, scan_desc)

        if "id" in scan_desc:
            concatted.attrs["id"] = scan_desc["id"]

        return concatted


class SingleFileEndstation(EndstationBase):
    """Abstract endstation which loads data from a single file.

    This just specializes the routine used to determine the location of files on disk.

    Unlike general endstations, if your data comes in a single file you can trust that the
    file given to you in the spreadsheet or direct load calls is all there is.
    """

    def resolve_frame_locations(self, scan_desc: dict = None):
        """Single file endstations just use the referenced file from the scan description."""
        if scan_desc is None:
            raise ValueError(
                "Must pass dictionary as file scan_desc to all endstation loading code."
            )

        original_data_loc = scan_desc.get("path", scan_desc.get("file"))
        p = Path(original_data_loc)
        if not p.exists():
            original_data_loc = os.path.join(
                getattr(arpes.config, "data_root", os.getcwd()), original_data_loc
            )

        p = Path(original_data_loc)
        return [p]


class HDF5Endstation(SingleFileEndstation):
    """
    Loads data from the HDF5 format. Assumes the data is stored following the conventions at beamline 7.
    """

    # If any other endstations use HDF5, we can try to generalize and move BL7 specific code to its plugin

    def load_single_frame(
        self, frame_path: str = None, scan_desc: dict = None, **kwargs
    ):
        """Loads a scan from a single .h5 file.

        This assumes the DAQ storage convention set by E. Rotenberg (possibly earlier authors)
        used at beamline 7.
        """
        hdf5 = h5py.File(frame_path, "r")

        all_coords, data_dimensions, scan_coord_names = construct_coords(hdf5)
        scan_shape = tuple(
            len(all_coords[coord_name]) for coord_name in scan_coord_names
        )

        data_vars = {}
        for dataset in [hdf5["0D_Data"], hdf5["1D_Data"], hdf5["2D_Data"]]:
            for data_name in dataset:
                if data_name in all_coords:
                    continue
                data = dataset_to_array(
                    dataset[data_name],
                    type="int32" if "Spectra" in data_name else "float64",
                )
                if len(scan_shape) > 1:
                    proper_shape = data.shape[:-1] + scan_shape
                    data = data.reshape(proper_shape)
                elif not scan_shape:
                    data = data.transpose()[0]
                coord_names = data_dimensions[data_name]
                data_vars[data_name] = xr.DataArray(
                    data,
                    coords={
                        coord_name: all_coords[coord_name] for coord_name in coord_names
                    },
                    dims=coord_names,
                    name=data_name,
                )

        attrs = get_attrs(hdf5)

        hdf5.close()

        return xr.Dataset(
            data_vars={
                f"{name}_safe" if name in all_coords else name: data
                for name, data in data_vars.items()
            },
            coords=all_coords,
            attrs=attrs,
        )

    def postprocess_frame(self, frame: xr.Dataset):
        frame = super().postprocess_frame(frame)

        deg_to_rad_coords = {"beta", "theta", "psi"}
        for coord_name in deg_to_rad_coords:
            for data_type in [frame.coords, frame.attrs]:
                if coord_name in data_type:
                    try:
                        data_type[coord_name] = data_type[coord_name] * (np.pi / 180)
                    except (TypeError, ValueError):
                        pass

        return frame


class FITSEndstation(EndstationBase):
    """Loads data from the .fits format from the MAESTRO software and derivatives.

    This ends up being somewhat complicated, because the FITS export is written in
    LabView and does not conform to the standard specification for the FITS archive
    format.

    Many of the intricacies here are in fact those shared between MAESTRO's format
    and the Lanzara Lab's format. Conrad does not foresee this as an issue, because it
    is unlikely that many other ARPES labs will adopt this data format moving forward,
    in light of better options derivative of HDF like the NeXuS format.
    """

    PREPPED_COLUMN_NAMES = {
        "time": "time",
        "Delay": "delay-var",  # these are named thus to avoid conflicts with the
        "Sample-X": "cycle-var",  # underlying coordinates
        "Mira": "pump_power",
        # insert more as needed
    }

    SKIP_COLUMN_NAMES = {
        "Phi",
        "null",
        "X",
        "Y",
        "Z",
        "mono_eV",
        "Slit Defl",
        "Optics Stage",
        "Scan X",
        "Scan Y",
        "Scan Z",
        # insert more as needed
    }

    SKIP_COLUMN_FORMULAS = {
        lambda name: True if ("beamview" in name or "IMAQdx" in name) else False,
    }

    RENAME_KEYS = {
        "Phi": "chi",
        "Beta": "beta",
        "Azimuth": "chi",
        "Pump_energy_uJcm2": "pump_fluence",
        "T0_ps": "t0_nominal",
        "W_func": "workfunction",
        "Slit": "slit",
        "LMOTOR0": "x",
        "LMOTOR1": "y",
        "LMOTOR2": "z",
        "LMOTOR3": "theta",
        "LMOTOR4": "beta",
        "LMOTOR5": "chi",
        "LMOTOR6": "alpha",
    }

    def resolve_frame_locations(self, scan_desc: dict = None):
        """These are stored as single files, so just use the one from the description."""
        if scan_desc is None:
            raise ValueError(
                "Must pass dictionary as file scan_desc to all endstation loading code."
            )

        original_data_loc = scan_desc.get("path", scan_desc.get("file"))
        p = Path(original_data_loc)
        if not p.exists():
            original_data_loc = os.path.join(arpes.config.data_root, original_data_loc)

        return [original_data_loc]

    def load_single_frame(
        self, frame_path: str = None, scan_desc: dict = None, **kwargs
    ):
        """Loads a scan from a single .fits file.

        This assumes the DAQ storage convention set by E. Rotenberg
        (possibly earlier authors) for the storage of ARPES data in FITS tables.

        This involves several complications:

        1. Hydrating/extracting coordinates from start/delta/n formats
        2. Extracting multiple scan regions
        3. Gracefully handling missing values
        4. Unwinding different scan conventions to common formats
        5. Handling early scan termination
        """
        # Use dimension labels instead of
        hdulist = fits.open(frame_path, ignore_missing_end=True)
        primary_dataset_name = None

        # Clean the header because sometimes out LabView produces improper FITS files
        for i in range(len(hdulist)):
            hdulist[i].header[
                "UN_0_0"
            ] = ""  # TODO This card is broken, this is not a good fix
            del hdulist[i].header["UN_0_0"]
            hdulist[i].header["UN_0_0"] = ""
            if "TTYPE2" in hdulist[i].header and hdulist[i].header["TTYPE2"] == "Delay":
                hdulist[i].header["TUNIT2"] = ""
                del hdulist[i].header["TUNIT2"]
                hdulist[i].header["TUNIT2"] = "ps"

            with warnings.catch_warnings():
                hdulist[i].verify("fix+warn")
                hdulist[i].header.update()
            # This actually requires substantially more work because it is lossy to
            # information on the unit that was encoded

        hdu = hdulist[1]

        scan_desc = copy.deepcopy(scan_desc)
        attrs = scan_desc.pop("note", scan_desc)
        attrs.update(dict(hdulist[0].header))

        drop_attrs = ["COMMENT", "HISTORY", "EXTEND", "SIMPLE", "SCANPAR", "SFKE_0"]
        for dropped_attr in drop_attrs:
            if dropped_attr in attrs:
                del attrs[dropped_attr]

        built_coords, dimensions, real_spectrum_shape = find_clean_coords(
            hdu, attrs, mode="MC"
        )

        # don't have phi because we need to convert pixels first
        deg_to_rad_coords = {"beta", "theta", "chi"}

        # convert angular attributes to radians
        for coord_name in deg_to_rad_coords:
            if coord_name in attrs:
                try:
                    attrs[coord_name] = float(attrs[coord_name]) * (np.pi / 180)
                except (TypeError, ValueError):
                    pass
            if coord_name in scan_desc:
                try:
                    scan_desc[coord_name] = float(scan_desc[coord_name]) * (np.pi / 180)
                except (TypeError, ValueError):
                    pass

        data_vars = {}

        all_names = hdu.columns.names
        n_spectra = len(
            [n for n in all_names if "Fixed_Spectra" in n or "Swept_Spectra" in n]
        )
        for column_name in hdu.columns.names:
            # we skip some fixed set of the columns, such as the one dimensional axes,
            # as well as things that are too
            # tricky to load at the moment, like the microscope images from MAESTRO
            should_skip = False
            if column_name in self.SKIP_COLUMN_NAMES:
                should_skip = True

            for formula in self.SKIP_COLUMN_FORMULAS:
                if formula(column_name):
                    should_skip = True

            if should_skip:
                continue

            # the hemisphere axis is handled below
            dimension_for_column = dimensions[column_name]
            column_shape = real_spectrum_shape[column_name]

            column_display = self.PREPPED_COLUMN_NAMES.get(column_name, column_name)
            if "Fixed_Spectra" in column_display:
                if n_spectra == 1:
                    column_display = "spectrum"
                else:
                    column_display = (
                        "spectrum" + "-" + column_display.split("Fixed_Spectra")[1]
                    )

            if "Swept_Spectra" in column_display:
                if n_spectra == 1:
                    column_display = "spectrum"
                else:
                    column_display = (
                        "spectrum" + "-" + column_display.split("Swept_Spectra")[1]
                    )

            # sometimes if a scan is terminated early it can happen that the sizes do not match the expected value
            # as an example, if a beta map is supposed to have 401 slices, it might end up having only 260 if it were
            # terminated early
            # If we are confident in our parsing code above, we can handle this case and take a subset of the coords
            # so that the data matches
            try:
                resized_data = hdu.data.columns[column_name].array.reshape(column_shape)
            except ValueError:
                # if we could not resize appropriately, we will try to reify the shapes together
                rest_column_shape = column_shape[1:]
                n_per_slice = int(np.prod(rest_column_shape))
                total_shape = hdu.data.columns[column_name].array.shape
                total_n = np.prod(total_shape)

                n_slices = total_n // n_per_slice
                # if this isn't true, we can't recover
                data_for_resize = hdu.data.columns[column_name].array
                if total_n // n_per_slice != total_n / n_per_slice:
                    # the last slice was in the middle of writing when something hit the fan
                    # we need to infer how much of the data to read, and then repeat the above
                    # we need to cut the data

                    # This can happen when the labview crashes during data collection,
                    # we use column_shape[1] because of the row order that is used in the FITS file
                    data_for_resize = data_for_resize[
                        0 : (total_n // n_per_slice) * column_shape[1]
                    ]
                    warnings.warn(
                        f"Column {column_name} was in the middle of slice when DAQ stopped. Throwing out incomplete slice..."
                    )

                column_shape = list(column_shape)
                column_shape[0] = n_slices

                try:
                    resized_data = data_for_resize.reshape(column_shape)
                except Exception:
                    # sometimes for whatever reason FITS errors and cannot read the data
                    continue

                # we also need to adjust the coordinates
                altered_dimension = dimension_for_column[0]
                built_coords[altered_dimension] = built_coords[altered_dimension][
                    :n_slices
                ]

            data_vars[column_display] = xr.DataArray(
                resized_data,
                coords={
                    k: c for k, c in built_coords.items() if k in dimension_for_column
                },
                dims=dimension_for_column,
            )

        # adjust angular coordinates
        built_coords = {
            k: c * (np.pi / 180) if k in deg_to_rad_coords else c
            for k, c in built_coords.items()
        }

        return xr.Dataset(
            {
                f"safe-{name}" if name in data_var.coords else name: data_var
                for name, data_var in data_vars.items()
            },
            attrs={**scan_desc, "name": primary_dataset_name},
        )


class SynchrotronEndstation(EndstationBase):
    """Base class code for ARPES setups at synchrotrons.

    Synchrotron endstations have somewhat complicated light source metadata.
    This stub exists to attach commonalities, such as a resolution table which
    can be interpolated into to retrieve the x-ray linewidth at the
    experimental settings. Additionally, subclassing this is used in resolution
    calculations to signal that such a resolution lookup is required.
    """

    RESOLUTION_TABLE = None


class HemisphericalEndstation(EndstationBase):
    """Base class code for ARPES setups using hemispheres.

    An endstation definition for a hemispherical analyzer should include
    everything needed to determine energy + k resolution, angle conversion,
    and ideally correction databases for dead pixels + detector nonlinearity
    information
    """

    ANALYZER_INFORMATION = None
    SLIT_ORIENTATION = None
    PIXELS_PER_DEG = None


def endstation_from_alias(alias: str) -> type:
    """Lookup the data loading class from an alias."""
    return _ENDSTATION_ALIASES[alias]


def endstation_name_from_alias(alias) -> str:
    """Lookup the data loading principal location from an alias."""
    return endstation_from_alias(alias).PRINCIPAL_NAME


def add_endstation(endstation_cls: type) -> None:
    """Registers a data loading plugin (Endstation class) together with its aliases.

    You can use this to add a plugin after the original search if it is defined in another
    module or in a notebook.
    """
    assert endstation_cls.PRINCIPAL_NAME is not None
    for alias in endstation_cls.ALIASES:
        if alias in _ENDSTATION_ALIASES:
            continue

        _ENDSTATION_ALIASES[alias] = endstation_cls

    _ENDSTATION_ALIASES[endstation_cls.PRINCIPAL_NAME] = endstation_cls


def resolve_endstation(retry=True, **kwargs) -> EndstationBase:
    """Tries to determine which plugin to use for loading a piece of data.

    Args:
        retry: Whether to attempt to reload plugins and try again after failure.
          This is used as an import guard basiscally in case the user imported things
          very strangely to ensure plugins are loaded.
        kwargs: Contains the actual information required to identify the scan.

    Returns:
        The loading plugin that should be used for the data.
    """
    endstation_name = case_insensitive_get(
        kwargs, "location", case_insensitive_get(kwargs, "endstation")
    )

    # check if the user actually provided a plugin
    if isinstance(endstation_name, type):
        return endstation_name

    if endstation_name is None:
        warnings.warn("Endstation not provided. Using `fallback` plugin.")
        endstation_name = "fallback"

    try:
        return endstation_from_alias(endstation_name)
    except KeyError:
        raise ValueError(
            "Could not identify endstation. "
            "Did you set the endstation or location? Find a description of the available options "
            "in the endstations module."
        )


def load_scan(scan_desc: dict[str, str], retry=True, **kwargs: Any) -> xr.Dataset:
    """Resolves a plugin and delegates loading a scan.

    This is used interally by `load_data` and should not be invoked directly
    by users.

    Determines which data loading class is appropriate for the data,
    shuffles a bit of metadata, and calls the .load function on the
    retrieved class to start the data loading process.

    Args:
        scan_desc: Information identifying the scan, typically a scan number or full path.
        retry: Used to attempt a reload of plugins and subsequent data load attempt.
        kwargs:

    Returns:
        Loaded and normalized ARPES scan data.
    """
    note = scan_desc.get("note", scan_desc)
    full_note = copy.deepcopy(scan_desc)
    full_note.update(note)

    endstation_cls = resolve_endstation(retry=retry, **full_note)

    key = "file" if "file" in scan_desc else "path"

    file = scan_desc[key]

    try:
        file = int(file)
        file = endstation_cls.find_first_file(file, scan_desc)
        scan_desc[key] = file
    except ValueError:
        pass

    endstation: EndstationBase = endstation_cls()
    return endstation.load(scan_desc, **kwargs)



"""Definitions of models involving Fermi edges."""

from lmfit.models import update_param_vals
from scipy import stats
import numpy as np
import lmfit as lf
from scipy.ndimage import gaussian_filter

from .x_model_mixin import XModelMixin
from .functional_forms import (
    fermi_dirac,
    fermi_dirac_affine,
    gstep,
    gstep_stdev,
    gstepb,
    lorentzian,
    band_edge_bkg,
    g,
    twolorentzian,
)

__all__ = [
    "AffineBroadenedFD",
    "FermiLorentzianModel",
    "FermiDiracModel",
    "GStepBModel",
    "TwoBandEdgeBModel",
    "BandEdgeBModel",
    "BandEdgeBGModel",
    "FermiDiracAffGaussModel",
    "GStepBStdevModel",
    "GStepBStandardModel",
    "TwoLorEdgeModel",
]


class AffineBroadenedFD(XModelMixin):
    """A model for fitting an affine density of states with resolution broadened Fermi-Dirac occupation."""

    @staticmethod
    def affine_broadened_fd(
        x, fd_center=0, fd_width=0.003, conv_width=0.02, const_bkg=1, lin_bkg=0, offset=0
    ):
        """Fermi function convoled with a Gaussian together with affine background.

        Args:
            x: value to evaluate function at
            fd_center: center of the step
            fd_width: width of the step
            conv_width: The convolution width
            const_bkg: constant background
            lin_bkg: linear background slope
            offset: constant background
        """
        dx = x - fd_center
        x_scaling = x[1] - x[0]
        fermi = 1 / (np.exp(dx / fd_width) + 1)
        return (
            gaussian_filter((const_bkg + lin_bkg * dx) * fermi, sigma=conv_width / x_scaling)
            + offset
        )

    def __init__(self, independent_vars=("x",), prefix="", missing="raise", name=None, **kwargs):
        """Defer to lmfit for initialization."""
        kwargs.update({"prefix": prefix, "missing": missing, "independent_vars": independent_vars})
        super().__init__(self.affine_broadened_fd, **kwargs)

        self.set_param_hint("offset", min=0.0)
        self.set_param_hint("fd_width", min=0.0)
        self.set_param_hint("conv_width", min=0.0)

    def guess(self, data, x=None, **kwargs):
        """Make some heuristic guesses.

        We use the mean value to estimate the background parameters and physically
        reasonable ones to initialize the edge.
        """
        pars = self.make_params()

        pars["%sfd_center" % self.prefix].set(value=0)
        pars["%slin_bkg" % self.prefix].set(value=0)
        pars["%sconst_bkg" % self.prefix].set(value=data.mean().item() * 2)
        pars["%soffset" % self.prefix].set(value=data.min().item())

        pars["%sfd_width" % self.prefix].set(0.005)  # TODO we can do better than this
        pars["%sconv_width" % self.prefix].set(0.02)

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC


class FermiLorentzianModel(XModelMixin):
    """A Lorentzian multiplied by a gstepb background."""

    @staticmethod
    def gstepb_mult_lorentzian(
        x, center=0, width=1, erf_amp=1, lin_bkg=0, const_bkg=0, gamma=1, lorcenter=0
    ):
        """A Lorentzian multiplied by a gstepb background."""
        return gstepb(x, center, width, erf_amp, lin_bkg, const_bkg) * lorentzian(
            x, gamma, lorcenter, 1
        )

    def __init__(self, independent_vars=("x",), prefix="", missing="raise", name=None, **kwargs):
        """Defer to lmfit for initialization."""
        kwargs.update({"prefix": prefix, "missing": missing, "independent_vars": independent_vars})
        super().__init__(self.gstepb_mult_lorentzian, **kwargs)

        self.set_param_hint("erf_amp", min=0.0)
        self.set_param_hint("width", min=0)
        self.set_param_hint("lin_bkg", min=-10, max=10)
        self.set_param_hint("const_bkg", min=-50, max=50)
        self.set_param_hint("gamma", min=0.0)

    def guess(self, data, x=None, **kwargs):
        """Placeholder for making better heuristic guesses here."""
        pars = self.make_params()

        pars["%scenter" % self.prefix].set(value=0)
        pars["%slorcenter" % self.prefix].set(value=0)
        pars["%slin_bkg" % self.prefix].set(value=0)
        pars["%sconst_bkg" % self.prefix].set(value=data.min())
        pars["%swidth" % self.prefix].set(0.02)  # TODO we can do better than this
        pars["%serf_amp" % self.prefix].set(value=data.mean() - data.min())

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC


class FermiDiracModel(XModelMixin):
    """A model for the Fermi Dirac function."""

    def __init__(self, independent_vars=("x",), prefix="", missing="drop", name=None, **kwargs):
        """Defer to lmfit for initialization."""
        kwargs.update({"prefix": prefix, "missing": missing, "independent_vars": independent_vars})
        super().__init__(fermi_dirac, **kwargs)

        self.set_param_hint("width", min=0)

    def guess(self, data, x=None, **kwargs):
        """Placeholder for making better heuristic guesses here."""
        pars = self.make_params()

        pars["{}center".format(self.prefix)].set(value=0)
        pars["{}width".format(self.prefix)].set(value=0.05)
        pars["{}scale".format(self.prefix)].set(value=data.mean() - data.min())

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC


class GStepBModel(XModelMixin):
    """A model for fitting Fermi functions with a linear background."""

    def __init__(self, independent_vars=("x",), prefix="", missing="raise", name=None, **kwargs):
        """Defer to lmfit for initialization."""
        kwargs.update({"prefix": prefix, "missing": missing, "independent_vars": independent_vars})
        super().__init__(gstepb, **kwargs)

        self.set_param_hint("erf_amp", min=0.0)
        self.set_param_hint("width", min=0)
        self.set_param_hint("lin_bkg", min=-10, max=10)
        self.set_param_hint("const_bkg", min=-50, max=50)

    def guess(self, data, x=None, **kwargs):
        """Placeholder for making better heuristic guesses here."""
        pars = self.make_params()

        pars["%scenter" % self.prefix].set(value=0)
        pars["%slin_bkg" % self.prefix].set(value=0)
        pars["%sconst_bkg" % self.prefix].set(value=data.min())
        pars["%swidth" % self.prefix].set(0.02)  # TODO we can do better than this
        pars["%serf_amp" % self.prefix].set(value=data.mean() - data.min())

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC


class TwoBandEdgeBModel(XModelMixin):
    """A model for fitting a Lorentzian and background multiplied into the fermi dirac distribution.

    TODO, actually implement two_band_edge_bkg (find original author and their intent).
    """

    @staticmethod
    def two_band_edge_bkg():
        """Some missing model referenced in old Igor code retained for visibility here."""
        raise NotImplementedError

    def __init__(self, independent_vars=("x",), prefix="", missing="raise", name=None, **kwargs):
        """Defer to lmfit for initialization."""
        kwargs.update(
            {
                "prefix": prefix,
                "missing": missing,
                "independent_vars": independent_vars,
            }
        )
        super().__init__(self.two_band_edge_bkg, **kwargs)

        self.set_param_hint("amplitude_1", min=0.0)
        self.set_param_hint("gamma_1", min=0.0)
        self.set_param_hint("amplitude_2", min=0.0)
        self.set_param_hint("gamma_2", min=0.0)

        self.set_param_hint("offset", min=-10)

    def guess(self, data, x=None, **kwargs):
        """Placeholder for making better heuristic guesses here.

        We should really do some peak fitting or edge detection to find
        okay values here.
        """
        pars = self.make_params()

        if x is not None:
            slope, intercept, r_value, p_value, std_err = stats.linregress(x, data)
            pars["%slor_center" % self.prefix].set(value=x[np.argmax(data - slope * x)])
        else:
            pars["%slor_center" % self.prefix].set(value=-0.2)

        pars["%sgamma" % self.prefix].set(value=0.2)
        pars["%samplitude" % self.prefix].set(value=(data.mean() - data.min()) / 1.5)

        pars["%sconst_bkg" % self.prefix].set(value=data.min())
        pars["%slin_bkg" % self.prefix].set(value=0)
        pars["%soffset" % self.prefix].set(value=data.min())

        pars["%scenter" % self.prefix].set(value=0)
        pars["%swidth" % self.prefix].set(0.02)  # TODO we can do better than this

        return update_param_vals(pars, self.prefix, **kwargs)


class BandEdgeBModel(XModelMixin):
    """A model for fitting a Lorentzian and background multiplied into the fermi dirac distribution."""

    def __init__(self, independent_vars=("x",), prefix="", missing="raise", name=None, **kwargs):
        """Defer to lmfit for initialization."""
        kwargs.update(
            {
                "prefix": prefix,
                "missing": missing,
                "independent_vars": independent_vars,
            }
        )
        super().__init__(band_edge_bkg, **kwargs)

        self.set_param_hint("amplitude", min=0.0)
        self.set_param_hint("gamma", min=0.0)
        self.set_param_hint("offset", min=-10)

    def guess(self, data, x=None, **kwargs):
        """Placeholder for making better heuristic guesses here.

        We should really do some peak fitting or edge detection to find
        okay values here.
        """
        pars = self.make_params()

        if x is not None:
            slope, intercept, r_value, p_value, std_err = stats.linregress(x, data)
            pars["%slor_center" % self.prefix].set(value=x[np.argmax(data - slope * x)])
        else:
            pars["%slor_center" % self.prefix].set(value=-0.2)

        pars["%sgamma" % self.prefix].set(value=0.2)
        pars["%samplitude" % self.prefix].set(value=(data.mean() - data.min()) / 1.5)

        pars["%sconst_bkg" % self.prefix].set(value=data.min())
        pars["%slin_bkg" % self.prefix].set(value=0)
        pars["%soffset" % self.prefix].set(value=data.min())

        pars["%scenter" % self.prefix].set(value=0)
        pars["%swidth" % self.prefix].set(0.02)  # TODO we can do better than this

        return update_param_vals(pars, self.prefix, **kwargs)


class BandEdgeBGModel(XModelMixin):
    """A model for fitting a Lorentzian and background multiplied into the fermi dirac distribution."""

    @staticmethod
    def band_edge_bkg_gauss(
        x,
        center=0,
        width=0.05,
        amplitude=1,
        gamma=0.1,
        lor_center=0,
        offset=0,
        lin_bkg=0,
        const_bkg=0,
    ):
        """A model for fitting a Lorentzian and background multiplied into the fermi dirac distribution."""
        return np.convolve(
            band_edge_bkg(x, 0, width, amplitude, gamma, lor_center, offset, lin_bkg, const_bkg),
            g(np.linspace(-6, 6, 800), 0, 0.01),
            mode="same",
        )

    def __init__(self, independent_vars=("x",), prefix="", missing="raise", name=None, **kwargs):
        """Defer to lmfit for initialization."""
        kwargs.update(
            {
                "prefix": prefix,
                "missing": missing,
                "independent_vars": independent_vars,
            }
        )
        super().__init__(self.band_edge_bkg_gauss, **kwargs)

        self.set_param_hint("amplitude", min=0.0)
        self.set_param_hint("gamma", min=0.0)
        self.set_param_hint("offset", min=-10)
        self.set_param_hint("center", vary=False)

    def guess(self, data, x=None, **kwargs):
        """Placeholder for making better heuristic guesses here.

        We should really do some peak fitting or edge detection to find
        okay values here.
        """
        pars = self.make_params()

        if x is not None:
            slope, intercept, r_value, p_value, std_err = stats.linregress(x, data)
            pars["%slor_center" % self.prefix].set(value=x[np.argmax(data - slope * x)])
        else:
            pars["%slor_center" % self.prefix].set(value=-0.2)

        pars["%sgamma" % self.prefix].set(value=0.2)
        pars["%samplitude" % self.prefix].set(value=(data.mean() - data.min()) / 1.5)

        pars["%sconst_bkg" % self.prefix].set(value=data.min())
        pars["%slin_bkg" % self.prefix].set(value=0)
        pars["%soffset" % self.prefix].set(value=data.min())

        # pars['%scenter' % self.prefix].set(value=0)
        pars["%swidth" % self.prefix].set(0.02)  # TODO we can do better than this

        return update_param_vals(pars, self.prefix, **kwargs)


class FermiDiracAffGaussModel(XModelMixin):
    """Fermi Dirac function with an affine background multiplied, then all convolved with a Gaussian."""

    @staticmethod
    def fermi_dirac_bkg_gauss(x, center=0, width=0.05, lin_bkg=0, const_bkg=0, scale=1, sigma=0.01):
        """Fermi Dirac function with an affine background multiplied, then all convolved with a Gaussian."""
        return np.convolve(
            fermi_dirac_affine(x, center, width, lin_bkg, const_bkg, scale),
            g(x, (min(x) + max(x)) / 2, sigma),
            mode="same",
        )

    def __init__(self, independent_vars=("x",), prefix="", missing="drop", name=None, **kwargs):
        """Defer to lmfit for initialization."""
        kwargs.update({"prefix": prefix, "missing": missing, "independent_vars": independent_vars})
        super().__init__(self.fermi_dirac_bkg_gauss, **kwargs)

        # self.set_param_hint('width', min=0)
        self.set_param_hint("width", vary=False)
        # self.set_param_hint('lin_bkg', max=10)
        # self.set_param_hint('scale', max=50000)
        self.set_param_hint("scale", min=0)
        self.set_param_hint("sigma", min=0, vary=True)
        self.set_param_hint("lin_bkg", vary=False)
        self.set_param_hint("const_bkg", vary=False)

    def guess(self, data, x=None, **kwargs):
        """Placeholder for making better heuristic guesses here."""
        pars = self.make_params()

        pars["{}center".format(self.prefix)].set(value=0)
        # pars['{}width'.format(self.prefix)].set(value=0.05)
        pars["{}width".format(self.prefix)].set(value=0.0009264)
        pars["{}scale".format(self.prefix)].set(value=data.mean() - data.min())
        pars["{}lin_bkg".format(self.prefix)].set(value=0)
        pars["{}const_bkg".format(self.prefix)].set(value=0)
        pars["{}sigma".format(self.prefix)].set(value=0.023)

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC


class GStepBStdevModel(XModelMixin):
    """A model for fitting Fermi functions with a linear background."""

    @staticmethod
    def gstepb_stdev(x, center=0, sigma=1, erf_amp=1, lin_bkg=0, const_bkg=0):
        """Fermi function convolved with a Gaussian together with affine background.

        Args:
            x: value to evaluate function at
            center: center of the step
            sigma: width of the step
            erf_amp: height of the step
            lin_bkg: linear background slope
            const_bkg: constant background
        """
        dx = x - center
        return const_bkg + lin_bkg * np.min(dx, 0) + gstep_stdev(x, center, sigma, erf_amp)

    def __init__(self, independent_vars=("x",), prefix="", missing="raise", name=None, **kwargs):
        """Defer to lmfit for initialization."""
        kwargs.update({"prefix": prefix, "missing": missing, "independent_vars": independent_vars})
        super().__init__(self.gstepb_stdev, **kwargs)

        self.set_param_hint("erf_amp", min=0.0)
        self.set_param_hint("sigma", min=0)
        self.set_param_hint("lin_bkg", min=-10, max=10)
        self.set_param_hint("const_bkg", min=-50, max=50)

    def guess(self, data, x=None, **kwargs):
        """Placeholder for making better heuristic guesses here."""
        pars = self.make_params()

        pars["%scenter" % self.prefix].set(value=0)
        pars["%slin_bkg" % self.prefix].set(value=0)
        pars["%sconst_bkg" % self.prefix].set(value=data.min())
        pars["%ssigma" % self.prefix].set(0.02)  # TODO we can do better than this
        pars["%serf_amp" % self.prefix].set(value=data.mean() - data.min())

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC


class GStepBStandardModel(XModelMixin):
    """A model for fitting Fermi functions with a linear background."""

    @staticmethod
    def gstepb_standard(x, center=0, sigma=1, amplitude=1, **kwargs):
        """Specializes paramters in gstepb."""
        return gstepb(x, center, width=sigma, erf_amp=amplitude, **kwargs)

    def __init__(self, independent_vars=("x",), prefix="", missing="raise", name=None, **kwargs):
        """Defer to lmfit for initialization."""
        kwargs.update({"prefix": prefix, "missing": missing, "independent_vars": independent_vars})
        super().__init__(self.gstepb_standard, **kwargs)

        self.set_param_hint("amplitude", min=0.0)
        self.set_param_hint("sigma", min=0)
        self.set_param_hint("lin_bkg", min=-10, max=10)
        self.set_param_hint("const_bkg", min=-50, max=50)

    def guess(self, data, x=None, **kwargs):
        """Placeholder for making better heuristic guesses here."""
        pars = self.make_params()

        pars["%scenter" % self.prefix].set(value=0)
        pars["%slin_bkg" % self.prefix].set(value=0)
        pars["%sconst_bkg" % self.prefix].set(value=data.min())
        pars["%ssigma" % self.prefix].set(0.02)  # TODO we can do better than this
        pars["%samplitude" % self.prefix].set(value=data.mean() - data.min())

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC


class TwoLorEdgeModel(XModelMixin):
    """A model for (two lorentzians with an affine background) multiplied by a gstepb."""

    def twolorentzian_gstep(
        x,
        gamma,
        t_gamma,
        center,
        t_center,
        amp,
        t_amp,
        lin_bkg,
        const_bkg,
        g_center,
        sigma,
        erf_amp,
    ):
        """Two Lorentzians, an affine background, and a gstepb edge."""
        TL = twolorentzian(x, gamma, t_gamma, center, t_center, amp, t_amp, lin_bkg, const_bkg)
        GS = gstep(x, g_center, sigma, erf_amp)
        return TL * GS

    def __init__(self, independent_vars=("x",), prefix="", missing="raise", name=None, **kwargs):
        """Defer to lmfit for initialization."""
        kwargs.update({"prefix": prefix, "missing": missing, "independent_vars": independent_vars})
        super().__init__(self.twolorentzian_gstep, **kwargs)

        self.set_param_hint("amp", min=0.0)
        self.set_param_hint("gamma", min=0)
        self.set_param_hint("t_amp", min=0.0)
        self.set_param_hint("t_gamma", min=0)
        self.set_param_hint("erf_amp", min=0.0)
        self.set_param_hint("sigma", min=0)
        self.set_param_hint("lin_bkg", min=-10, max=10)
        self.set_param_hint("const_bkg", min=-50, max=50)

    def guess(self, data, x=None, **kwargs):
        """Placeholder for making better heuristic guesses here."""
        pars = self.make_params()

        pars["%scenter" % self.prefix].set(value=0)
        pars["%st_center" % self.prefix].set(value=0)
        pars["%sg_center" % self.prefix].set(value=0)
        pars["%slin_bkg" % self.prefix].set(value=0)
        pars["%sconst_bkg" % self.prefix].set(value=data.min())
        pars["%sgamma" % self.prefix].set(0.02)  # TODO we can do better than this
        pars["%st_gamma" % self.prefix].set(0.02)
        pars["%ssigma" % self.prefix].set(0.02)
        pars["%samp" % self.prefix].set(value=data.mean() - data.min())
        pars["%st_amp" % self.prefix].set(value=data.mean() - data.min())
        pars["%serf_amp" % self.prefix].set(value=data.mean() - data.min())

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC



"""Some utilities for optics and optical design.

This and the utilities in arpes.laser should maybe be grouped into a separate place.

We don't do any astigmatism aware calculations here, which might be of practical utility.

Things to offer in the future:

1. Calculating beam characteristics (beam divergence, focus, etc) from a sequence of
   knife edge tests or beam profiler images in order to facilitate common tasks like beam
   collimation, focusing, or preparing a new Tr-ARPES setup.
2. Nonlinear optics utilities including damage thresholds to allow simpler design of
   harmonic generation for Tr-ARPES.
"""

# pylint: disable=invalid-name,no-value-for-parameter

import numpy as np

__all__ = (
    "waist",
    "waist_R",
    "rayleigh_range",
    "lens_transfer",
    "magnification",
    "waist_from_divergence",
    "waist_from_rr",
)


def waist(wavelength: float, z: float, z_R: float) -> float:
    """Calculates the waist size from the measurements at a distance from the waist."""
    raise NotImplementedError


def waist_R(waist_0: float, m_squared: float = 1.0) -> float:
    """Calculates the width of t he beam a distance from the waist."""
    raise NotImplementedError
    # return np.sqrt(m_squared) * waist()


def waist_from_rr(wavelength: float, rayleigh_rng: float) -> float:
    """Calculates the waist parameters from the Rayleigh range."""
    return np.sqrt((wavelength * rayleigh_rng) / np.pi)


def rayleigh_range(wavelength: float, beam_waist: float, m_squared: float = 1.0) -> float:
    """Calculates the Rayleigh range from beam parameters at the waist."""
    return np.pi * (beam_waist ** 2) / (m_squared * wavelength)


def lens_transfer(s: float, f: float, beam_rayleigh_range: float, m_squared: float = 1.0) -> float:
    """Lens transfer calculation. Calculates the location of the image.

    Args:
        s: The object distance.
        f: The focal length of the lens.
        beam_rayleigh_range: The Rayleigh range for the beam, uncorrected for the beam quality factor.
        m_squared: The beam quality factor.

    Returns:
        The distance to the image.
    """
    t = 1 / f - 1 / (s + (beam_rayleigh_range / m_squared) ** 2 / (s - f))
    return 1 / t


def waist_from_divergence(wavelength: float, half_angle_divergence: float) -> float:
    """Calculates the waist from the beam's half angle divergence and wavelength."""
    return wavelength / (np.pi * half_angle_divergence)


def magnification(s: float, f: float, beam_rayleigh_range: float, m_squared: float = 1.0) -> float:
    """Calculates the magnification offered by a lens system.

    Args:
        s: The object distance.
        f: The focal length of the lens.
        beam_rayleigh_range: The Rayleigh range for the beam, uncorrected for the beam quality factor.
        m_squared: The beam quality factor.

    Returns:
        The magnificiation of a lens.
    """
    return 1 / np.sqrt((1 - s / f) ** 2 + (beam_rayleigh_range / f / m_squared) ** 2)



"""Imports momentum conversion routines for forward and inverse (volumetric) conversion."""
from .core import *
from .remap_manipulator import *
from .forward import *
from .calibration import *
from .trapezoid import *



"""A widget providing controls for a coordinate offset in the momentum tool."""
from functools import partial

from PyQt5 import QtWidgets

__all__ = ("CoordinateOffsetWidget",)


class CoordinateOffsetWidget(QtWidgets.QGroupBox):
    """Control for a single coordinate offset in the momentum tool."""

    def __init__(self, parent=None, root=None, coordinate_name=None, value=None):
        """Configures utility label, an inner control, and a linked spinbox for text entry."""
        super().__init__(title=coordinate_name, parent=parent)

        self.layout = QtWidgets.QGridLayout(self)

        self.label = QtWidgets.QLabel("Value: ")
        self.spinbox = QtWidgets.QSpinBox()
        self.slider = QtWidgets.QSlider()
        self.root = root

        self.layout.addWidget(self.label)
        self.layout.addWidget(self.spinbox)
        self.layout.addWidget(self.slider)

        self._prevent_change_events = False

        self.slider.valueChanged.connect(partial(self.value_changed, source=self.slider))
        self.spinbox.valueChanged.connect(partial(self.value_changed, source=self.spinbox))

        self.recompute()

    def value_changed(self, event, source):
        """Propagates values change to update the displayed data."""
        if self._prevent_change_events:
            return

        self._prevent_change_events = True
        self.slider.setValue(source.value())
        self.spinbox.setValue(source.value())
        self._prevent_change_events = False
        self.recompute()
        if self.root is not None:
            self.root.update_cut()

    def recompute(self):
        """Recompute stale UI state from this control's value."""
        value = self.spinbox.value()
        self.label.setText("Value: {:.3g}".format(value))

    def value(self):
        """The value of the control, which just forwards the inner widget."""
        return self.spinbox.value()



"""Implements Igor <-> xarray interop, notably loading Igor waves and packed experiment files."""

import warnings
from pathlib import Path

import xarray as xr

from arpes.utilities import rename_keys, safe_decode

from pygor import load, Wave


__all__ = ("read_single_pxt",)


def read_header(header_bytes: bytes):
    header_as_string = safe_decode(header_bytes)

    lines = [x for x in header_as_string.replace("\r", "\n").split("\n") if x]
    lines = [x for x in lines if "=" in x]

    header = {}
    for line in lines:
        fragments = line.split("=")
        first, rest = fragments[0], "=".join(fragments[1:])

        try:
            rest = int(rest)
        except ValueError:
            try:
                rest = float(rest)
            except ValueError:
                pass

        header[first.lower().replace(" ", "_")] = rest

    return rename_keys(
        header,
        {
            "sample_x": "x",
            "sample_y_(vert)": "y",
            "sample_y": "y",
            "sample_z": "z",
            "bl_energy": "hv",
        },
    )


def wave_to_xarray(wave: Wave) -> xr.DataArray:
    """Convert a wave to an `xr.DataArray`.

    Units, if present on the wave, are used to furnish the dimension names. If dimension
    names are not present, placeholder names ("X", "Y", "Z", "W", as in Igor) are used
    for each unitless dimension.

    Args:
        wave: The input wave, an `igor.Wave` instance.

    Returns:
        The converted `xr.DataArray` instance.
    """
    # only need four because Igor only supports four dimensions!
    extra_names = iter(["W", "X", "Y", "Z"])
    n_dims = len([a for a in wave.axis if len(a)])

    def get_axis_name(index: int) -> str:
        unit = wave.axis_units[index]
        if unit:
            return {
                "eV": "eV",
                "deg": "phi",
                "Pwr Supply V": "volts",
                "K2200 V": "volts",
            }.get(unit, unit)

        return next(extra_names)

    axis_names = [get_axis_name(i) for i in range(n_dims)]
    coords = dict(zip(axis_names, wave.axis))

    return xr.DataArray(
        wave.data,
        coords=coords,
        dims=axis_names,
        attrs=read_header(wave.notes),
    )


def read_single_pxt(
    reference_path: Path | str,
) -> xr.DataArray:
    """Use igor.igorpy to load a single .PXT or .PXP file."""

    if isinstance(reference_path, Path):
        reference_path = str(reference_path.absolute())

    loaded = None
    for try_byte_order in [">", "=", "<"]:
        try:
            loaded = load(reference_path, initial_byte_order=try_byte_order)
            break
        except Exception:  # pylint: disable=broad-except
            # bad byte ordering probably
            pass

    children = [c for c in loaded.children if isinstance(c, Wave)]

    if len(children) == 1:
        return wave_to_xarray(children[0])

    warnings.warn(
        f"Igor PXT file contained {len(children)} waves. Ignoring all but first."
    )
    return wave_to_xarray(children[0])



"""Utilities for selecting and defining masks on data interactively."""
import numpy as np

from arpes.analysis.mask import apply_mask
from arpes.exceptions import AnalysisError
from arpes.plotting.interactive_utils import CursorTool, SaveableTool
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum

__all__ = ["mask"]


class MaskTool(SaveableTool, CursorTool):
    """Tool to allow masking data by drawing regions."""

    auto_zero_nans = False
    auto_rebin = False

    def __init__(self, **kwargs):
        super().__init__(kwargs.pop("name", None))

        self.load_settings(**kwargs)

        self.app_main_size = self.settings.get("app_main_size", 600)
        self.app_marginal_size = self.settings.get("app_main_size", 300)
        self.pointer_mode = "region"

    def tool_handler(self, doc):
        from bokeh import events
        from bokeh.layouts import row, column
        from bokeh.models.mappers import LinearColorMapper
        from bokeh.models import widgets
        from bokeh.plotting import figure

        if len(self.arr.shape) != 2:
            raise AnalysisError("Cannot use mask tool on non image-like spectra")

        arr = self.arr
        x_coords, y_coords = arr.coords[arr.dims[0]], arr.coords[arr.dims[1]]

        default_palette = self.default_palette

        self.app_context.update(
            {
                "region_options": [],
                "regions": {},
                "data": arr,
                "data_range": {
                    "x": (np.min(x_coords.values), np.max(x_coords.values)),
                    "y": (np.min(y_coords.values), np.max(y_coords.values)),
                },
            }
        )

        self.cursor = [np.mean(self.data_range["x"]), np.mean(self.data_range["y"])]

        self.color_maps["main"] = LinearColorMapper(
            default_palette, low=np.min(arr.values), high=np.max(arr.values), nan_color="black"
        )

        main_tools = ["wheel_zoom", "tap", "reset"]
        main_title = "Mask Tool: WARNING Unidentified"

        try:
            main_title = "Mask Tool: {}".format(arr.S.label[:60])
        except:
            pass

        self.figures["main"] = figure(
            tools=main_tools,
            plot_width=self.app_main_size,
            plot_height=self.app_main_size,
            min_border=10,
            min_border_left=20,
            toolbar_location="left",
            x_axis_location="below",
            y_axis_location="right",
            title=main_title,
            x_range=self.data_range["x"],
            y_range=self.data_range["y"],
        )

        self.figures["main"].xaxis.axis_label = arr.dims[0]
        self.figures["main"].yaxis.axis_label = arr.dims[1]

        self.plots["main"] = self.figures["main"].image(
            [np.asarray(arr.values.T)],
            x=self.data_range["x"][0],
            y=self.data_range["y"][0],
            dw=self.data_range["x"][1] - self.data_range["x"][0],
            dh=self.data_range["y"][1] - self.data_range["y"][0],
            color_mapper=self.color_maps["main"],
        )

        self.add_cursor_lines(self.figures["main"])
        region_patches = self.figures["main"].patches(
            xs=[], ys=[], color="white", alpha=0.35, line_width=1
        )

        def add_point_to_region():
            if self.active_region in self.regions:
                self.regions[self.active_region]["points"].append(list(self.cursor))
                update_region_display()

            self.save_app()

        def click_main_image(event):
            self.cursor = [event.x, event.y]
            if self.pointer_mode == "region":
                add_point_to_region()

        POINTER_MODES = [
            (
                "Cursor",
                "cursor",
            ),
            (
                "Region",
                "region",
            ),
        ]

        def perform_mask(data=None, **kwargs):
            if data is None:
                data = arr

            data = normalize_to_spectrum(data)
            return apply_mask(data, self.app_context["mask"], **kwargs)

        self.app_context["perform_mask"] = perform_mask
        self.app_context["mask"] = None

        pointer_dropdown = widgets.Dropdown(
            label="Pointer Mode", button_type="primary", menu=POINTER_MODES
        )
        self.region_dropdown = widgets.Dropdown(
            label="Active Region", button_type="primary", menu=self.region_options
        )

        edge_mask_button = widgets.Button(label="Edge Mask")
        region_name_input = widgets.TextInput(placeholder="Region name...")
        add_region_button = widgets.Button(label="Add Region")

        clear_region_button = widgets.Button(label="Clear Region")
        remove_region_button = widgets.Button(label="Remove Region")

        main_color_range_slider = widgets.RangeSlider(
            start=0,
            end=100,
            value=(
                0,
                100,
            ),
            title="Color Range",
        )

        def on_click_edge_mask():
            if self.active_region in self.regions:
                old_points = self.regions[self.active_region]["points"]
                dims = [d for d in arr.dims if "eV" != d]
                energy_index = arr.dims.index("eV")
                max_energy = np.max([p[energy_index] for p in old_points])

                other_dim = dims[0]
                other_coord = arr.coords[other_dim].values
                min_other, max_other = np.min(other_coord), np.max(other_coord)
                min_e = np.min(arr.coords["eV"].values)

                if arr.dims.index("eV") == 0:
                    before = [[min_e - 3, min_other - 1], [0, min_other - 1]]
                    after = [[0, max_other + 1], [min_e - 3, max_other + 1]]
                else:
                    before = [[min_other - 1, min_e - 3], [min_other - 1, 0]]
                    after = [[max_other + 1, 0], [max_other + 1, min_e - 3]]
                self.regions[self.active_region]["points"] = before + old_points + after
                self.app_context["mask"] = self.app_context["mask"] or {}
                self.app_context["mask"]["fermi"] = max_energy
                update_region_display()

            self.save_app()

        def add_region(region_name):
            if region_name not in self.regions:
                self.region_options.append(
                    (
                        region_name,
                        region_name,
                    )
                )
                self.region_dropdown.menu = self.region_options
                self.regions[region_name] = {
                    "points": [],
                    "name": region_name,
                }

                if self.active_region is None:
                    self.active_region = region_name

                self.save_app()

        def on_change_active_region(event):
            region_id = event.item
            self.app_context["active_region"] = region_id
            self.active_region = region_id
            self.save_app()

        def on_change_pointer_mode(event):
            pointer_mode = event.item
            self.app_context["pointer_mode"] = pointer_mode
            self.pointer_mode = pointer_mode
            self.save_app()

        def update_region_display():
            region_names = self.regions.keys()

            if self.app_context["mask"] is None:
                self.app_context["mask"] = {}
            self.app_context["mask"].update(
                {"dims": arr.dims, "polys": [r["points"] for r in self.regions.values()]}
            )

            region_patches.data_source.data = {
                "xs": [[p[0] for p in self.regions[r]["points"]] for r in region_names],
                "ys": [[p[1] for p in self.regions[r]["points"]] for r in region_names],
            }
            self.save_app()

        self.update_region_display = update_region_display

        def on_clear_region():
            if self.active_region in self.regions:
                self.regions[self.active_region]["points"] = []
                update_region_display()

        def on_remove_region():
            if self.active_region in self.regions:
                del self.regions[self.active_region]
                new_region_options = [b for b in self.region_options if b[0] != self.active_region]
                self.region_dropdown.menu = new_region_options
                self.region_options = new_region_options
                self.active_region = None
                update_region_display()

        # Attach callbacks
        main_color_range_slider.on_change("value", self.update_colormap_for("main"))

        self.figures["main"].on_event(events.Tap, click_main_image)
        self.region_dropdown.on_click(on_change_active_region)
        pointer_dropdown.on_click(on_change_pointer_mode)
        add_region_button.on_click(lambda: add_region(region_name_input.value))
        edge_mask_button.on_click(on_click_edge_mask)
        clear_region_button.on_click(on_clear_region)
        remove_region_button.on_click(on_remove_region)

        layout = row(
            column(self.figures["main"]),
            column(
                *[
                    f
                    for f in [
                        column(
                            pointer_dropdown,
                            self.region_dropdown,
                        ),
                        row(
                            region_name_input,
                            add_region_button,
                        ),
                        edge_mask_button if "eV" in arr.dims else None,
                        row(
                            clear_region_button,
                            remove_region_button,
                        ),
                        column(
                            self._cursor_info,
                            main_color_range_slider,
                        ),
                    ]
                    if f is not None
                ]
            ),
        )

        doc.add_root(layout)
        doc.title = "Mask Tool"
        self.load_app()
        self.save_app()

    def serialize(self):
        return {
            "active_region": self.active_region,
            "region_options": self.region_options,
            "regions": self.regions,
            "cursor": self.cursor,
        }

    def deserialize(self, json_data):
        self.cursor = json_data.get("cursor", [0, 0])

        self.app_context["regions"] = json_data.get("regions", {}) or {}
        self.app_context["region_options"] = json_data.get("region_options", [])

        self.region_dropdown.menu = self.app_context["region_options"]

        self.update_region_display()


def mask(data: DataType, **kwargs):
    """Start an interactive mask selection tool."""
    data = normalize_to_spectrum(data)

    tool = MaskTool(**kwargs)
    return tool.make_tool(data)



"""Contains many common utility functions for managing matplotlib."""

import collections
import pickle
import contextlib
from typing import List, Tuple, Union

import datetime
import re
import itertools
import json
import os.path
import warnings
import pathlib
from collections import Counter

import matplotlib
import matplotlib.cm as cm
import matplotlib.offsetbox
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import colorbar, colors, gridspec
from matplotlib.lines import Line2D

import xarray as xr
from arpes import VERSION
from arpes.settings import SETTINGS
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum
from arpes.utilities.jupyter import get_recent_history, get_notebook_name

__all__ = (
    # General + IO
    "path_for_plot",
    "path_for_holoviews",
    "name_for_dim",
    "unit_for_dim",
    "load_data_for_figure",
    "savefig",
    "AnchoredHScaleBar",
    "calculate_aspect_ratio",
    # context managers
    "dark_background",
    # color related
    "temperature_colormap",
    "polarization_colorbar",
    "temperature_colormap_around",
    "temperature_colorbar",
    "temperature_colorbar_around",
    "generic_colorbarmap",
    "generic_colorbarmap_for_data",
    "colorbarmaps_for_axis",
    # Axis generation
    "dos_axes",
    "simple_ax_grid",
    # matplotlib 'macros'
    "invisible_axes",
    "no_ticks",
    "get_colorbars",
    "remove_colorbars",
    "frame_with",
    "unchanged_limits",
    "imshow_arr",
    "imshow_mask",
    "lineplot_arr",  # 1D version of imshow_arr
    "plot_arr",  # generic dimension version of imshow_arr, plot_arr
    # insets related
    "inset_cut_locator",
    "swap_xaxis_side",
    "swap_yaxis_side",
    "swap_axis_sides",
    # units related
    "data_to_axis_units",
    "axis_to_data_units",
    "daxis_ddata_units",
    "ddata_daxis_units",
    # TeX related
    "quick_tex",
    "latex_escape",
    # Decorating + labeling
    "label_for_colorbar",
    "label_for_dim",
    "label_for_symmetry_point",
    "sum_annotation",
    "mean_annotation",
    "fancy_labels",
    "mod_plot_to_ax",
    # Data summaries
    "summarize",
    "transform_labels",
    "v_gradient_fill",
    "h_gradient_fill",
)


@contextlib.contextmanager
def unchanged_limits(ax):
    """Context manager that retains axis limits."""
    xlim, ylim = ax.get_xlim(), ax.get_ylim()

    yield

    ax.set_xlim(xlim)
    ax.set_ylim(ylim)


def mod_plot_to_ax(data, ax, mod, **kwargs):
    """Plots a model onto an axis using the data range from the passed data."""
    with unchanged_limits(ax):
        xs = data.coords[data.dims[0]].values
        ys = mod.eval(x=xs)
        ax.plot(xs, ys, **kwargs)


def h_gradient_fill(
    x1, x2, x_solid, fill_color=None, ax=None, zorder=None, alpha=None, **kwargs
):
    """Fills a gradient between x1 and x2.

    If x_solid is not None, the gradient will be extended
    at the maximum opacity from the closer limit towards x_solid.

    Args:
        x1
        x2
        x_solid
        fill_color
        ax
        zorder
        alpha
        **kwargs

    Returns:
        The result of the inner imshow.
    """
    if ax is None:
        ax = plt.gca()

    xlim, ylim = ax.get_xlim(), ax.get_ylim()
    assert fill_color

    alpha = 1.0 if alpha is None else alpha

    z = np.empty((1, 100, 4), dtype=float)

    rgb = matplotlib.colors.colorConverter.to_rgb(fill_color)
    z[:, :, :3] = rgb
    z[:, :, -1] = np.linspace(0, alpha, 100)[None, :]

    xmin, xmax, (ymin, ymax) = x1, x2, ylim
    im = ax.imshow(
        z, aspect="auto", extent=[xmin, xmax, ymin, ymax], origin="lower", zorder=zorder
    )

    if x_solid is not None:
        xlow, xhigh = (x2, x_solid) if x_solid > x2 else (x_solid, x1)
        ax.fill_betweenx(ylim, xlow, xhigh, color=fill_color, alpha=alpha)

    ax.set_xlim(xlim)
    ax.set_ylim(ylim)
    return im


def v_gradient_fill(
    y1, y2, y_solid, fill_color=None, ax=None, zorder=None, alpha=None, **kwargs
):
    """Fills a gradient vertically between y1 and y2.

    If y_solid is not None, the gradient will be extended
    at the maximum opacity from the closer limit towards y_solid.

    Args:
        y1
        y2
        y_solid
        fill_color
        ax
        zorder
        alpha
        **kwargs

    Returns:
        The result of the inner imshow call.
    """
    if ax is None:
        ax = plt.gca()

    xlim, ylim = ax.get_xlim(), ax.get_ylim()
    assert fill_color

    alpha = 1.0 if alpha is None else alpha

    z = np.empty((100, 1, 4), dtype=float)

    rgb = matplotlib.colors.colorConverter.to_rgb(fill_color)
    z[:, :, :3] = rgb
    z[:, :, -1] = np.linspace(0, alpha, 100)[:, None]

    (xmin, xmax), ymin, ymax = xlim, y1, y2
    im = ax.imshow(
        z, aspect="auto", extent=[xmin, xmax, ymin, ymax], origin="lower", zorder=zorder
    )

    if y_solid is not None:
        ylow, yhigh = (y2, y_solid) if y_solid > y2 else (y_solid, y1)
        ax.fill_between(xlim, ylow, yhigh, color=fill_color, alpha=alpha)

    ax.set_xlim(xlim)
    ax.set_ylim(ylim)
    return im


def simple_ax_grid(
    n_axes, figsize=None, **kwargs
) -> Tuple[plt.Figure, List[plt.Axes], List[plt.Axes]]:
    """Generates a square-ish set of axes and hides the extra ones.

    It would be nice to accept an "aspect ratio" item that will attempt to fix the
    grid dimensions to get an aspect ratio close to the desired one.

    Args:
        n_axes
        figsize
        **kwargs

    Returns:
        The figure, the first n axis which are shown, and the remaining hidden axes.
    """
    width = int(np.ceil(np.sqrt(n_axes)))
    height = width - 1
    if width * height < n_axes:
        height += 1

    if figsize is None:
        figsize = (
            3 * max(width, 5),
            3 * max(height, 5),
        )

    fig, ax = plt.subplots(height, width, figsize=figsize, **kwargs)
    if n_axes == 1:
        ax = np.array([ax])

    ax, ax_rest = ax.ravel()[:n_axes], ax.ravel()[n_axes:]
    for axi in ax_rest:
        invisible_axes(axi)

    return fig, ax, ax_rest


@contextlib.contextmanager
def dark_background(overrides):
    """Context manager for plotting "dark mode"."""
    defaults = {
        "axes.edgecolor": "white",
        "xtick.color": "white",
        "ytick.color": "white",
        "axes.labelcolor": "white",
        "text.color": "white",
    }
    defaults.update(overrides)

    with plt.rc_context(defaults):
        yield


def data_to_axis_units(points, ax=None):
    """Converts between data and axis units."""
    if ax is None:
        ax = plt.gca()

    return ax.transAxes.inverted().transform(ax.transData.transform(points))


def axis_to_data_units(points, ax=None):
    """Converts between axis and data units."""
    if ax is None:
        ax = plt.gca()

    return ax.transData.inverted().transform(ax.transAxes.transform(points))


def ddata_daxis_units(ax=None):
    """Gives the derivative of data units with respect to axis units."""
    if ax is None:
        ax = plt.gca()

    dp1 = axis_to_data_units((1.0, 1.0), ax)
    dp0 = axis_to_data_units((0.0, 0.0), ax)
    return dp1 - dp0


def daxis_ddata_units(ax=None):
    """Gives the derivative of axis units with respect to data units."""
    if ax is None:
        ax = plt.gca()

    dp1 = data_to_axis_units((1.0, 1.0), ax)
    dp0 = data_to_axis_units((0.0, 0.0), ax)
    return dp1 - dp0


def swap_xaxis_side(ax):
    """Swaps the x axis to the top of the figure."""
    ax.xaxis.tick_top()
    ax.xaxis.set_label_position("top")


def swap_yaxis_side(ax):
    """Swaps the y axis to the right of the figure."""
    ax.yaxis.tick_right()
    ax.yaxis.set_label_position("right")


def swap_axis_sides(ax):
    """Swaps the y axis to the right of the figure and the x axis to the top."""
    swap_xaxis_side(ax)
    swap_yaxis_side(ax)


def transform_labels(transform_fn, fig=None, include_titles=True):
    """Apply a function to all axis labeled in a figure."""
    if fig is None:
        fig = plt.gcf()

    axes = list(fig.get_axes())
    for ax in axes:
        try:
            ax.set_xlabel(transform_fn(ax.get_xlabel(), is_title=False))
            ax.set_ylabel(transform_fn(ax.get_xlabel(), is_title=False))
            if include_titles:
                ax.set_title(transform_fn(ax.get_title(), is_title=True))
        except TypeError:
            ax.set_xlabel(transform_fn(ax.get_xlabel()))
            ax.set_ylabel(transform_fn(ax.get_xlabel()))
            if include_titles:
                ax.set_title(transform_fn(ax.get_title()))


def summarize(data: DataType, axes=None):
    """Makes a summary plot with different marginal plots represented."""
    data = normalize_to_spectrum(data)

    axes_shapes_for_dims = {
        1: (1, 1),
        2: (1, 1),
        3: (2, 2),  # one extra here
        4: (3, 2),  # corresponds to 4 choose 2 axes
    }

    if axes is None:
        fig, axes = plt.subplots(
            axes_shapes_for_dims.get(len(data.dims)), figsize=(8, 8)
        )

    flat_axes = axes.ravel()
    combinations = list(itertools.combinations(data.dims, 2))
    for axi, combination in zip(flat_axes, combinations):
        data.sum(combination).plot(ax=axi)
        fancy_labels(axi)

    for i in range(len(combinations), len(flat_axes)):
        flat_axes[i].set_axis_off()

    return axes


def sum_annotation(eV=None, phi=None):
    """Annotates that a given axis was summed over by listing the integration range."""
    eV_annotation, phi_annotation = "", ""

    def to_str(bound):
        if bound is None:
            return ""

        return "{:.2f}".format(bound)

    if eV is not None:
        if SETTINGS.get("use_tex", False):
            eV_annotation = (
                "$\\text{E}_{" + to_str(eV.start) + "}^{" + to_str(eV.stop) + "}$"
            )
        else:
            eV_annotation = to_str(eV.start) + " < E < " + to_str(eV.stop)
    if phi is not None:
        if SETTINGS.get("use_tex", False):
            phi_annotation = (
                "$\\phi_{" + to_str(phi.start) + "}^{" + to_str(phi.stop) + "}$"
            )
        else:
            phi_annotation = to_str(phi.start) + " < φ < " + to_str(phi.stop)

    return eV_annotation + phi_annotation


def mean_annotation(eV=None, phi=None):
    """Annotates that a given axis was meaned (summed) over by listing the integration range."""
    eV_annotation, phi_annotation = "", ""

    def to_str(bound):
        if bound is None:
            return ""

        return "{:.2f}".format(bound)

    if eV is not None:
        if SETTINGS.get("use_tex", False):
            eV_annotation = (
                "$\\bar{\\text{E}}_{"
                + to_str(eV.start)
                + "}^{"
                + to_str(eV.stop)
                + "}$"
            )
        else:
            eV_annotation = (
                "Mean<" + to_str(eV.start) + " < E < " + to_str(eV.stop) + ">"
            )
    if phi is not None:
        if SETTINGS.get("use_tex", False):
            phi_annotation = (
                "$\\bar{\\phi}_{" + to_str(phi.start) + "}^{" + to_str(phi.stop) + "}$"
            )
        else:
            phi_annotation = (
                "Mean<" + to_str(phi.start) + " < φ < " + to_str(phi.stop) + ">"
            )

    return eV_annotation + phi_annotation


def frame_with(ax, color="red", linewidth=2):
    """Makes thick, visually striking borders on a matplotlib plot.

    Very useful for color coding results in a slideshow.
    """
    for spine in ["left", "right", "top", "bottom"]:
        ax.spines[spine].set_color(color)
        ax.spines[spine].set_linewidth(linewidth)


LATEX_ESCAPE_MAP = {
    "_": r"\_",
    "<": r"\textless{}",
    ">": r"\textgreater{}",
    "{": r"\{",
    "}": r"\}",
    "&": r"\&",
    "%": r"\%",
    "$": r"\$",
    "#": r"\#",
    "~": r"\textasciitilde{}",
    "^": r"\^{}",
    "\\": r"\textbackslash{}",
}
LATEX_ESCAPE_REGEX = re.compile(
    "|".join(
        re.escape(str(k))
        for k in sorted(LATEX_ESCAPE_MAP.keys(), key=lambda item: -len(item))
    )
)


def latex_escape(text: str, force: bool = False) -> str:
    """Conditionally escapes a string based on the matplotlib settings.

    If you need the escaped string even if you are not using matplotlib with LaTeX
    support, you can pass `force=True`.

    Adjusted from suggestions at:
    https://stackoverflow.com/questions/16259923/how-can-i-escape-latex-special-characters-inside-django-templates

    Args:
        text: The contents which should be escaped
        force: Whether we should perform escaping even if matplotlib is
          not being used with LaTeX support.

    Returns:
        The escaped string which should appear in LaTeX with the same
        contents as the original.
    """
    if not matplotlib.rcParams["text.usetex"] and not force:
        return text

    # otherwise, we need to escape
    return LATEX_ESCAPE_REGEX.sub(lambda match: LATEX_ESCAPE_MAP[match.group()], text)


def quick_tex(latex_fragment: str, ax=None, fontsize=30) -> plt.Axes:
    """Sometimes you just need to render some LaTeX.

    Getting a LaTex session running is far too much effort.
    Also just go to the KaTeX website and can work well.

    Args:
        latex_fragment: The fragment to render

    Returns:
        The axes generated.
    """
    if ax is None:
        fig, ax = plt.subplots()

    invisible_axes(ax)
    ax.text(0.2, 0.2, latex_fragment, fontsize=fontsize)
    return ax


def lineplot_arr(arr, ax=None, method="plot", mask=None, mask_kwargs=None, **kwargs):
    """Convenience method to plot an array with a mask over some other data."""
    if mask_kwargs is None:
        mask_kwargs = dict()

    if ax is None:
        _, ax = plt.subplots()

    xs = None
    if arr is not None:
        fn = plt.plot
        if method == "scatter":
            fn = plt.scatter

        xs = arr.coords[arr.dims[0]].values
        fn(xs, arr.values, **kwargs)

    if mask is not None:
        y_lim = ax.get_ylim()
        if isinstance(mask, list) and isinstance(mask[0], slice):
            for slice_mask in mask:
                ax.fill_betweenx(
                    y_lim, slice_mask.start, slice_mask.stop, **mask_kwargs
                )
        else:
            raise NotImplementedError
        ax.set_ylim(y_lim)

    return ax


def plot_arr(arr=None, ax=None, over=None, mask=None, **kwargs):
    """Convenience method to plot an array with a mask over some other data."""
    to_plot = arr if mask is None else mask
    try:
        n_dims = len(to_plot.dims)
    except AttributeError:
        n_dims = 1

    if n_dims == 2:
        quad = None
        if arr is not None:
            ax, quad = imshow_arr(arr, ax=ax, over=over, **kwargs)
        if mask is not None:
            over = quad if over is None else over
            imshow_mask(mask, ax=ax, over=over, **kwargs)
    if n_dims == 1:
        ax = lineplot_arr(arr, ax=ax, mask=mask, **kwargs)

    return ax


def imshow_mask(mask, ax=None, over=None, cmap=None, **kwargs):
    """Plots a mask by using a fixed color and transparency."""
    assert over is not None

    if ax is None:
        ax = plt.gca()

    if cmap is None:
        cmap = "Reds"

    if isinstance(cmap, str):
        cmap = cm.get_cmap(name=cmap)

    cmap.set_bad("k", alpha=0)

    ax.imshow(
        mask.values,
        cmap=cmap,
        interpolation="none",
        vmax=1,
        vmin=0,
        origin="lower",
        extent=over.get_extent(),
        aspect=ax.get_aspect(),
        **kwargs,
    )


def imshow_arr(
    arr,
    ax=None,
    over=None,
    origin="lower",
    aspect="auto",
    alpha=None,
    vmin=None,
    vmax=None,
    cmap=None,
    **kwargs,
):
    """Similar to plt.imshow but users different default origin, and sets appropriate extents.

    Args:
        arr
        ax

    Returns:
        The axes and quadmesh instance.
    """
    if ax is None:
        fig, ax = plt.subplots()

    x, y = arr.coords[arr.dims[0]].values, arr.coords[arr.dims[1]].values
    extent = [y[0], y[-1], x[0], x[-1]]

    if over is None:
        if alpha is not None:
            if vmin is None:
                vmin = arr.min().item()
            if vmax is None:
                vmax = arr.max().item()
            if cmap is None:
                cmap = "viridis"
            if isinstance(cmap, str):
                cmap = cm.get_cmap(cmap)
            norm = colors.Normalize(vmin=vmin, vmax=vmax)
            mappable = cm.ScalarMappable(cmap=cmap, norm=norm)
            mapped_colors = mappable.to_rgba(arr.values)
            mapped_colors[:, :, 3] = alpha
            quad = ax.imshow(
                mapped_colors, origin=origin, extent=extent, aspect=aspect, **kwargs
            )
        else:
            quad = ax.imshow(
                arr.values,
                origin=origin,
                extent=extent,
                aspect=aspect,
                cmap=cmap,
                **kwargs,
            )
        ax.grid(False)
        ax.set_xlabel(arr.dims[1])
        ax.set_ylabel(arr.dims[0])
    else:
        quad = ax.imshow(
            arr.values,
            extent=over.get_extent(),
            aspect=ax.get_aspect(),
            origin=origin,
            **kwargs,
        )

    return ax, quad


def dos_axes(
    orientation="horiz", figsize=None, with_cbar=True
) -> Tuple[plt.Figure, plt.Axes]:
    """Makes axes corresponding to density of states data.

    This has one image like region and one small marginal for an EDC.
    Orientation option should be 'horiz' or 'vert'.

    Args:
        orientation
        figsize
        with_cbar

    Returns:
        The generated figure and axes as a tuple.
    """
    if figsize is None:
        figsize = (12, 9) if orientation == "vert" else (9, 9)

    fig = plt.figure(figsize=figsize)

    outer_grid = gridspec.GridSpec(4, 4, wspace=0.0, hspace=0.0)

    if orientation == "horiz":
        fig.subplots_adjust(hspace=0.00)
        gs = gridspec.GridSpec(2, 1, height_ratios=[3, 1])

        ax0 = plt.subplot(gs[0])
        axes = (ax0, plt.subplot(gs[1], sharex=ax0))
        plt.setp(axes[0].get_xticklabels(), visible=False)
    else:
        fig.subplots_adjust(wspace=0.00)
        gs = gridspec.GridSpec(1, 2, width_ratios=[1, 4])

        ax0 = plt.subplot(gs[1])
        axes = (ax0, plt.subplot(gs[0], sharey=ax0))
        plt.setp(axes[0].get_yticklabels(), visible=False)

    return fig, axes


def inset_cut_locator(
    data, reference_data=None, ax=None, location=None, color=None, **kwargs
):
    """Plots a reference cut location over a figure.

    Another approach is to separately plot the locator and add it in Illustrator or
    another tool.

    Args:
        data: The data you are plotting
        reference_data: The reference data containing the location of the cut
        ax: The axes to plot on
        location: The location in the cut
        color: The color to use for the indicator line
        kwargs: Passed to ax.plot when making the indicator lines
    """
    quad = data.plot(ax=ax)
    ax.set_xlabel("")
    ax.set_ylabel("")
    try:
        quad.colorbar.remove()
    except Exception:
        pass

    # add more as necessary
    missing_dim_resolvers = {
        "theta": lambda: reference_data.S.theta,
        "beta": lambda: reference_data.S.beta,
        "phi": lambda: reference_data.S.phi,
    }

    missing_dims = [d for d in data.dims if d not in location]
    missing_values = {d: missing_dim_resolvers[d]() for d in missing_dims}
    ordered_selector = [location.get(d, missing_values.get(d)) for d in data.dims]

    n = 200

    def resolve(name, value):
        if isinstance(value, slice):
            low = value.start
            high = value.stop

            if low is None:
                low = data.coords[name].min().item()
            if high is None:
                high = data.coords[name].max().item()

            return np.linspace(low, high, n)

        return np.ones((n,)) * value

    n_cut_dims = len(
        [d for d in ordered_selector if isinstance(d, (collections.Iterable, slice))]
    )
    ordered_selector = [resolve(d, v) for d, v in zip(data.dims, ordered_selector)]

    if missing_dims:
        assert reference_data is not None
        print(missing_dims)

    if n_cut_dims == 2:
        # a region cut, illustrate with a rect or by suppressing background
        return
    if color is None:
        color = "red"

    if n_cut_dims == 1:
        # a line cut, illustrate with a line
        ax.plot(*ordered_selector[::-1], color=color, **kwargs)
    elif n_cut_dims == 0:
        # a single point cut, illustrate with a marker
        pass


def generic_colormap(low, high):
    """Generates a colormap from the cm.Blues palette, suitable for most purposes."""
    delta = high - low
    low = low - delta / 6
    high = high + delta / 6

    def get_color(value):
        return cm.Blues(float((value - low) / (high - low)))

    return get_color


def phase_angle_colormap(low=0, high=np.pi * 2):
    """Generates a colormap suitable for angular data or data on a unit circle like a phase."""

    def get_color(value):
        return cm.twilight_shifted(float((value - low) / (high - low)))

    return get_color


def delay_colormap(low=-1, high=1):
    """Generates a colormap suitable for pump-probe delay data."""

    def get_color(value):
        return cm.coolwarm(float((value - low) / (high - low)))

    return get_color


def temperature_colormap(high=300, low=0, cmap=None):
    """Generates a colormap suitable for temperature data with fixed extent."""
    if cmap is None:
        cmap = cm.Blues_r

    def get_color(value):
        return cmap(float((value - low) / (high - low)))

    return get_color


def temperature_colormap_around(central, range=50):
    """Generates a colormap suitable for temperature data around a central value."""

    def get_color(value):
        return cm.RdBu_r(float((value - central) / range))

    return get_color


def generic_colorbar(low, high, label="", cmap=None, ax=None, ticks=None, **kwargs):
    extra_kwargs = {
        "orientation": "horizontal",
        "label": label,
        "ticks": ticks if ticks is not None else [low, high],
    }

    delta = high - low
    low = low - delta / 6
    high = high + delta / 6

    extra_kwargs.update(kwargs)
    cb = colorbar.ColorbarBase(
        ax,
        cmap=cm.get_cmap(cmap or "Blues"),
        norm=colors.Normalize(vmin=low, vmax=high),
        **extra_kwargs,
    )

    return cb


def phase_angle_colorbar(high=np.pi * 2, low=0, ax=None, **kwargs):
    """Generates a colorbar suitable for plotting an angle or value on a unit circle."""
    extra_kwargs = {
        "orientation": "horizontal",
        "label": "Angle",
        "ticks": ["0", r"$\pi$", r"$2\pi$"],
    }

    if not SETTINGS.get("use_tex", False):
        extra_kwargs["ticks"] = ["0", "π", "2π"]

    extra_kwargs.update(kwargs)
    cb = colorbar.ColorbarBase(
        ax,
        cmap=cm.get_cmap("twilight_shifted"),
        norm=colors.Normalize(vmin=low, vmax=high),
        **extra_kwargs,
    )
    return cb


def temperature_colorbar(high=300, low=0, ax=None, cmap=None, **kwargs):
    """Generates a colorbar suitable for temperature data with fixed extent."""
    if cmap is None:
        cmap = "Blues_r"

    extra_kwargs = {
        "orientation": "horizontal",
        "label": "Temperature (K)",
        "ticks": [low, high],
    }
    extra_kwargs.update(kwargs)
    cb = colorbar.ColorbarBase(
        ax, cmap=cmap, norm=colors.Normalize(vmin=low, vmax=high), **extra_kwargs
    )
    return cb


def delay_colorbar(low=-1, high=1, ax=None, **kwargs):
    """Generates a colorbar suitable for delay data.

    TODO make this nonsequential for use in case where you want to have a long time period after the
    delay or before.
    """
    extra_kwargs = {
        "orientation": "horizontal",
        "label": "Probe Pulse Delay (ps)",
        "ticks": [low, 0, high],
    }
    extra_kwargs.update(kwargs)
    cb = colorbar.ColorbarBase(
        ax, cmap="coolwarm", norm=colors.Normalize(vmin=low, vmax=high), **extra_kwargs
    )
    return cb


def temperature_colorbar_around(central, range=50, ax=None, **kwargs):
    """Generates a colorbar suitable for temperature axes around a central value."""
    extra_kwargs = {
        "orientation": "horizontal",
        "label": "Temperature (K)",
        "ticks": [central - range, central + range],
    }
    extra_kwargs.update(kwargs)
    cb = colorbar.ColorbarBase(
        ax,
        cmap="RdBu_r",
        norm=colors.Normalize(vmin=central - range, vmax=central + range),
        **extra_kwargs,
    )
    return cb


colorbarmaps_for_axis = {
    "temp": (
        temperature_colorbar,
        temperature_colormap,
    ),
    "delay": (
        delay_colorbar,
        delay_colormap,
    ),
    "theta": (
        phase_angle_colorbar,
        phase_angle_colormap,
    ),
    "volts": (
        generic_colorbar,
        generic_colormap,
    ),
}


def get_colorbars(fig=None) -> List[plt.Axes]:
    """Collects likely colorbars in a figure."""
    if fig is None:
        fig = plt.gcf()

    colorbars = []
    for ax in fig.axes:
        if ax.get_aspect() == 20:
            colorbars.append(ax)

    return colorbars


def remove_colorbars(fig=None):
    """Removes colorbars from given (or, if no given figure, current) matplotlib figure.

    Args:
        fig: The figure to modify, by default uses the current figure (`plt.gcf()`)
    """
    # TODO after colorbar removal, plots should be relaxed/rescaled to occupy space previously allocated to colorbars
    # for now, can follow this with plt.tight_layout()
    try:
        if fig is not None:
            for ax in fig.axes:
                if ax.get_aspect() == 20:  # a bit of a hack
                    ax.remove()
        else:
            remove_colorbars(plt.gcf())
    except Exception:
        pass


generic_colorbarmap = (
    generic_colorbar,
    generic_colormap,
)


def generic_colorbarmap_for_data(
    data: xr.DataArray, keep_ticks=True, ax=None, **kwargs
):
    """Generates a colorbar and colormap which is useful in general context."""
    low, high = data.min().item(), data.max().item()
    ticks = None
    if keep_ticks:
        ticks = data.values
    return (
        generic_colorbar(low=low, high=high, ax=ax, ticks=kwargs.get("ticks", ticks)),
        generic_colormap(low=low, high=high),
    )


def polarization_colorbar(ax=None):
    """Makes a colorbar which is appropriate for "polarization" (e.g. spin) data."""
    cb = colorbar.ColorbarBase(
        ax,
        cmap="RdBu",
        norm=colors.Normalize(vmin=-1, vmax=1),
        orientation="horizontal",
        label="Polarization",
        ticks=[-1, 0, 1],
    )
    return cb


def calculate_aspect_ratio(data: DataType):
    """Calculate the aspect ratio which should be used for plotting some data based on extent."""
    data = normalize_to_spectrum(data)

    assert len(data.dims) == 2

    x_extent = np.ptp(data.coords[data.dims[0]].values)
    y_extent = np.ptp(data.coords[data.dims[1]].values)

    return y_extent / x_extent


class AnchoredHScaleBar(matplotlib.offsetbox.AnchoredOffsetbox):
    """Provides an anchored scale bar on the X axis.

    Modified from `this StackOverflow question <https://stackoverflow.com/questions/43258638/>`_
    as alternate to the one provided through matplotlib.
    """

    def __init__(
        self,
        size=1,
        extent=0.03,
        label="",
        loc=2,
        ax=None,
        pad=0.4,
        borderpad=0.5,
        ppad=0,
        sep=2,
        prop=None,
        label_color=None,
        frameon=True,
        **kwargs,
    ):
        """Setup the scale bar and coordinate transforms to the parent axis."""
        if not ax:
            ax = plt.gca()
        trans = ax.get_xaxis_transform()

        size_bar = matplotlib.offsetbox.AuxTransformBox(trans)
        line = Line2D([0, size], [0, 0], **kwargs)
        vline1 = Line2D([0, 0], [-extent / 2.0, extent / 2.0], **kwargs)
        vline2 = Line2D([size, size], [-extent / 2.0, extent / 2.0], **kwargs)
        size_bar.add_artist(line)
        size_bar.add_artist(vline1)
        size_bar.add_artist(vline2)
        txt = matplotlib.offsetbox.TextArea(
            label,
            minimumdescent=False,
            textprops={
                "color": label_color,
            },
        )
        self.vpac = matplotlib.offsetbox.VPacker(
            children=[size_bar, txt], align="center", pad=ppad, sep=sep
        )
        matplotlib.offsetbox.AnchoredOffsetbox.__init__(
            self,
            loc,
            pad=pad,
            borderpad=borderpad,
            child=self.vpac,
            prop=prop,
            frameon=frameon,
        )


def load_data_for_figure(p: Union[str, pathlib.Path]):
    """Tries to load the data associated with a given figure by unpickling the saved data."""
    path = str(p)
    stem = os.path.splitext(path)[0]
    if stem.endswith("-PAPER"):
        stem = stem[:-6]

    pickle_file = stem + ".pickle"

    if not os.path.exists(pickle_file):
        raise ValueError("No saved data matching figure.")

    with open(pickle_file, "rb") as f:
        data = pickle.load(f)

    return data


def path_for_plot(desired_path):
    raise NotImplementedError("This function is not yet implemented.")


def savefig(desired_path, dpi=400, data=None, save_data=None, paper=False, **kwargs):
    """The PyARPES preferred figure saving routine.

    Provides a number of conveniences over matplotlib's `savefig`:

    #. Output is scoped per project and per day, which aids organization
    #. The dpi is set to a reasonable value for the year 2021.
    #. By omitting a file extension you will get high and low res formats in .png and .pdf
       which is useful for figure drafting in external software (Adobe Illustrator)
    #. Data and plot provenenace is tracked, which makes it easier to find your analysis
       after the fact if you have many many plots.

    """
    if not os.path.splitext(desired_path)[1]:
        paper = True

    if save_data is None:
        if paper:
            raise ValueError(
                "You must supply save_data when outputting in paper mode. This "
                "is for your own good so you can more easily regenerate "
                "the figure later!"
            )
    else:
        output_location = path_for_plot(os.path.splitext(desired_path)[0])
        with open(output_location + ".pickle", "wb") as f:
            pickle.dump(save_data, f)

    if paper:
        # automatically generate useful file formats
        high_dpi = max(dpi, 400)
        formats_for_paper = [
            "pdf",
            "png",
        ]  # not including SVG anymore because files too large

        for format in formats_for_paper:
            savefig(
                f"{desired_path}-PAPER.{format}",
                dpi=high_dpi,
                data=data,
                paper=False,
                **kwargs,
            )

        savefig(
            f"{desired_path}-low-PAPER.pdf", dpi=200, data=data, paper=False, **kwargs
        )

        return

    full_path = path_for_plot(desired_path)
    provenance_path = full_path + ".provenance.json"
    provenance_context = {
        "VERSION": VERSION,
        "time": datetime.datetime.now().isoformat(),
        "jupyter_notebook_name": get_notebook_name(),
        "name": "savefig",
    }

    def extract(for_data):
        try:
            return for_data.attrs.get("provenance", {})
        except Exception:
            return {}

    if data is not None:
        assert isinstance(
            data,
            (
                list,
                tuple,
                set,
            ),
        )
        provenance_context.update(
            {
                "jupyter_context": get_recent_history(1),
                "data": [extract(d) for d in data],
            }
        )
    else:
        # get more recent history because we don't have the data
        provenance_context.update(
            {
                "jupyter_context": get_recent_history(5),
            }
        )

    with open(provenance_path, "w") as f:
        json.dump(provenance_context, f, indent=2)
    plt.savefig(full_path, dpi=dpi, **kwargs)


def path_for_holoviews(desired_path):
    """Determines an appropriate output path for a holoviews save."""
    skip_paths = [".svg", ".png", ".jpeg", ".jpg", ".gif"]

    prefix, ext = os.path.splitext(desired_path)

    if ext in skip_paths:
        return prefix

    return prefix + ext


def name_for_dim(dim_name, escaped=True):
    """Alternate variant of `label_for_dim`."""
    if SETTINGS.get("use_tex", False):
        name = {
            "temperature": "Temperature",
            "beta": r"$\beta$",
            "theta": r"$\theta$",
            "chi": r"$\chi$",
            "alpha": r"$\alpha$",
            "psi": r"$\psi$",
            "phi": r"$\phi",
            "eV": r"$\textnormal{E}$",
            "kx": r"$\textnormal{k}_\textnormal{x}$",
            "ky": r"$\textnormal{k}_\textnormal{y}$",
            "kz": r"$\textnormal{k}_\textnormal{z}$",
            "kp": r"$\textnormal{k}_\textnormal{\parallel}$",
            "hv": r"$h\nu$",
        }.get(dim_name)
    else:
        name = {
            "temperature": "Temperature",
            "beta": "β",
            "theta": "θ",
            "chi": "χ",
            "alpha": "α",
            "psi": "ψ",
            "phi": "φ",
            "eV": "E",
            "kx": "Kx",
            "ky": "Ky",
            "kz": "Kz",
            "kp": "Kp",
            "hv": "Photon Energy",
        }.get(dim_name)

    if not escaped:
        name = name.replace("$", "")

    return name


def unit_for_dim(dim_name, escaped=True):
    """Calculate LaTeX or fancy display label for the unit associated to a dimension."""
    if SETTINGS.get("use_tex", False):
        unit = {
            "temperature": "K",
            "theta": r"rad",
            "beta": r"rad",
            "psi": r"rad",
            "chi": r"rad",
            "alpha": r"rad",
            "phi": r"rad",
            "eV": r"eV",
            "kx": r"$\AA^{-1}$",
            "ky": r"$\AA^{-1}$",
            "kz": r"$\AA^{-1}$",
            "kp": r"$\AA^{-1}$",
            "hv": r"eV",
        }.get(dim_name)
    else:
        unit = {
            "temperature": "K",
            "theta": r"rad",
            "beta": r"rad",
            "psi": r"rad",
            "chi": r"rad",
            "alpha": r"rad",
            "phi": r"rad",
            "eV": r"eV",
            "kx": "1/Å",
            "ky": "1/Å",
            "kz": "1/Å",
            "kp": "1/Å",
            "hv": "eV",
        }.get(dim_name)

    if not escaped:
        unit = unit.replace("$", "")

    return unit


def label_for_colorbar(data):
    """Returns an appropriate label for an ARPES intensity colorbar."""
    if not data.S.is_differentiated:
        return r"Spectrum Intensity (arb.)"

    # determine which axis was differentiated
    hist = data.S.history
    records = [h["record"] for h in hist if isinstance(h, dict)]
    if "curvature" in [r["by"] for r in records]:
        curvature_record = [r for r in records if r["by"] == "curvature"][0]
        directions = curvature_record["directions"]
        return r"Curvature along {} and {}".format(
            name_for_dim(directions[0]), name_for_dim(directions[1])
        )

    derivative_records = [r for r in records if r["by"] == "dn_along_axis"]
    c = Counter(
        itertools.chain(*[[d["axis"]] * d["order"] for d in derivative_records])
    )

    partial_frag = r""
    if sum(c.values()) > 1:
        partial_frag = r"^" + str(sum(c.values()))

    return (
        r"$\frac{\partial"
        + partial_frag
        + r" \textnormal{Int.}}{"
        + r"".join(
            [
                r"\partial {}^{}".format(name_for_dim(item, escaped=False), n)
                for item, n in c.items()
            ]
        )
        + "}$ (arb.)"
    )


def label_for_dim(data=None, dim_name=None, escaped=True):
    """Generates a fancy label (LaTeX, if available) for a dimension according to standard conventions."""
    if SETTINGS.get("use_tex", False):
        raw_dim_names = {
            "temperature": "Temperature",
            "theta": r"$\theta$",
            "beta": r"$\beta$",
            "chi": r"$\chi$",
            "alpha": r"$\alpha$",
            "psi": r"$\psi$",
            "phi": r"$\varphi$",
            "eV": r"Binding Energy (eV)",
            "angle": r"Interp. Angle",
            "kinetic": r"Kinetic Energy (eV)",
            "temp": r"Temperature",
            "kp": r"$k_\parallel$",
            "kx": r"$k_\text{x}$",
            "ky": r"$k_\text{y}$",
            "kz": r"$k_\perp$",
            "hv": "Photon Energy",
            "x": "X (mm)",
            "y": "Y (mm)",
            "z": "Z (mm)",
            "spectrum": "Intensity (arb.)",
        }
    else:
        raw_dim_names = {
            "temperature": "Temperature",
            "beta": "β",
            "theta": "θ",
            "chi": "χ",
            "alpha": "α",
            "psi": "ψ",
            "phi": "φ",
            "eV": "Binding Energy (eV)",
            "angle": "Interp. Angle",
            "kinetic": "Kinetic Energy (eV)",
            "temp": "Temperature (K)",
            "kp": "Kp",
            "kx": "Kx",
            "ky": "Ky",
            "kz": "Kz",
            "hv": "Photon Energy (eV)",
            "x": "X (mm)",
            "y": "Y (mm)",
            "z": "Z (mm)",
            "spectrum": "Intensity (arb.)",
        }

    if dim_name in raw_dim_names:
        return raw_dim_names.get(dim_name)

    try:
        from titlecase import titlecase
    except ImportError:
        warnings.warn(
            "Using alternative titlecase, for better results `pip install titlecase`."
        )

        def titlecase(s: str) -> str:
            """Poor man's titlecase.

            Args:
                s: The input string

            Returns:
                The titlecased string.
            """
            return s.title()

    result = titlecase(dim_name.replace("_", " "))
    return result


def fancy_labels(ax_or_ax_set, data=None):
    """Attaches better display axis labels for all axes.

    Axes are determined by those that can be traversed in the passed figure or axes.

    Args:
        ax_or_ax_set: The axis to search for subaxes
        data: The source data, used to calculate names, typically you can leave this empty
    """
    if isinstance(ax_or_ax_set, (list, tuple, set, np.ndarray)):
        for ax in ax_or_ax_set:
            fancy_labels(ax)
        return

    ax = ax_or_ax_set
    try:
        ax.set_xlabel(label_for_dim(data=data, dim_name=ax.get_xlabel()))
    except Exception as e:
        raise e
        pass

    try:
        ax.set_ylabel(label_for_dim(data=data, dim_name=ax.get_ylabel()))
    except Exception:
        pass


def label_for_symmetry_point(point_name: str) -> str:
    """Determines the LaTeX label for a symmetry point shortcode."""
    if SETTINGS.get("use_tex", False):
        proper_names = {
            "G": r"$\Gamma$",
            "X": r"X",
            "Y": r"Y",
        }
    else:
        proper_names = {
            "G": r"Γ",
            "X": r"X",
            "Y": r"Y",
        }

    return proper_names.get(point_name, point_name)


class CoincidentLinesPlot:
    """Helper to allow drawing lines at the same location.

    Will draw n lines offset so that their center appears at the data center,
    and the lines will end up nonoverlapping.

    Only works for straight lines.

    Technique adapted from `StackOverflow
    <https://stackoverflow.com/questions/19394505/matplotlib-expand-the-line-with-specified-width-in-data-unit>`_.
    """

    linewidth = 3

    def __init__(self, **kwargs):
        self.ax = kwargs.pop("ax", plt.gca())
        self.fig = kwargs.pop("fig", plt.gcf())
        self.extra_kwargs = kwargs
        self.ppd = 72.0 / self.fig.dpi
        self.has_drawn = False

        self.events = {
            "resize_event": self.ax.figure.canvas.mpl_connect(
                "resize_event", self._resize
            ),
            "motion_notify_event": self.ax.figure.canvas.mpl_connect(
                "motion_notify_event", self._resize
            ),
            "button_release_event": self.ax.figure.canvas.mpl_connect(
                "button_release_event", self._resize
            ),
        }
        self.handles = []
        self.lines = (
            []
        )  # saved args and kwargs for plotting, does not verify coincidence

    def add_line(self, *args, **kwargs):
        """Adds an additional line into the collection to be drawn."""
        assert not self.has_drawn
        self.lines.append(
            (
                args,
                kwargs,
            )
        )

    def draw(self):
        """Draw all of the lines after offsetting them slightly."""
        self.has_drawn = True

        offset_in_data_units = self.data_units_per_pixel * self.linewidth
        self.offsets = [
            offset_in_data_units * (o - (len(self.lines) - 1) / 2)
            for o in range(len(self.lines))
        ]

        for offset, (line_args, line_kwargs) in zip(self.offsets, self.lines):
            line_args = self.normalize_line_args(line_args)
            line_args[1] = np.array(line_args[1]) + offset
            handle = self.ax.plot(*line_args, **line_kwargs)
            self.handles.append(handle)

    @property
    def data_units_per_pixel(self):
        """Gets the data/pixel conversion ratio."""
        trans = self.ax.transData.transform
        inverse = (trans((1, 1)) - trans((0, 0))) * self.ppd
        return (1 / inverse[0], 1 / inverse[1])

    def normalize_line_args(self, args):
        def is_data_type(value):
            return isinstance(value, (np.array, np.ndarray, list, tuple))

        assert is_data_type(args[0])

        if len(args) > 1 and is_data_type(args[1]) and len(args[0]) == len(args[1]):
            # looks like we have x and y data
            return args

        # otherwise we should pad the args with the x data
        return [range(len(args[0]))] + args

    def _resize(self, event=None):
        # Keep the trace in here until we can test appropriately.
        import pdb

        pdb.set_trace()
        """
        self.line.set_linewidth(lw)
        self.ax.figure.canvas.draw_idle()
        self.lw = lw
        """


def invisible_axes(ax):
    """Make a Axes instance completely invisible."""
    ax.grid(False)
    ax.set_axis_off()
    ax.patch.set_alpha(0)


def no_ticks(ax):
    """Remove all axis ticks."""
    ax.get_xaxis().set_ticks([])
    ax.get_yaxis().set_ticks([])



"""Utility functions for extracting ARPES information from the HDF5 file conventions."""

import numpy as np
import h5py
from .common import safe_reshape

__all__ = (
    "construct_coords",
    "get_attrs",
    "dataset_to_array",
)


def get_scan_coords(
    scan_info: dict[str, str], scalar_data: h5py.Dataset
) -> dict[str, np.ndarray]:
    """Gets the scan coordinates from the scan information in the headers"""

    n_loops = int(scan_info["LWLVLPN"])
    scan_coords = {}
    for loop in range(n_loops):
        n_scan_coords = int(scan_info[f"NMSBDV{loop}"])
        n_scan_dimensions = 0
        for i in range(n_scan_coords):
            if f"ST_{loop}_{i}" in scan_info:
                n_scan_dimensions += 1

        shape = tuple(
            [int(scan_info[f"N_{loop}_{i}"]) for i in range(n_scan_dimensions)][::-1]
        )

        for scan_dimension in range(n_scan_dimensions):
            if (f"ST_{loop}_{scan_dimension}" not in scan_info) or (
                shape[scan_dimension] == 1
            ):
                continue
            name = scan_info[f"NM_{loop}_{scan_dimension}"]
            raw_data = dataset_to_array(scalar_data[name])
            reshaped_data = np.moveaxis(
                safe_reshape(raw_data, shape), scan_dimension, 0
            )
            averaged_data = reshaped_data.mean(axis=tuple(range(n_scan_dimensions - 1)))
            scan_coords[name] = averaged_data

    return scan_coords


def construct_coords(
    hdf5: h5py.File,
) -> tuple[dict[str, np.ndarray], dict[str, tuple[str]], tuple[str]]:
    """
    Constructs all coordinates from the HDF5 file, including the scan coordinates and the detector coordinates.
    Returns a dictionary of the coordinates and a dictionary of the dimensions for each data variable
    """

    scan_header = hdf5["Headers"]["Scan"] if "Scan" in hdf5["Headers"] else []
    low_level_scan_header = hdf5["Headers"]["Low_Level_Scan"]
    scan_info = {}
    for item in list(scan_header) + list(low_level_scan_header):
        item: list[bytes]
        try:
            scan_info[item[1].decode("utf-8").strip()] = (
                item[2].decode("utf-8").replace("'", "")
            )
        except UnicodeDecodeError:
            pass

    scan_coords = get_scan_coords(scan_info, hdf5["0D_Data"])
    scan_coord_names = tuple(scan_coords.keys())
    data_dimensions = {}
    constructed_coords = scan_coords.copy()

    for scalar_data in hdf5["0D_Data"]:
        if scalar_data in scan_coords:
            continue
        data_dimensions[scalar_data] = tuple(scan_coords.keys())

    for n_dims in (1, 2):
        dataset = hdf5[f"{n_dims}D_Data"]
        for data_name in dataset:
            attrs = dataset[data_name].attrs
            offsets = attrs["scaleOffset"][::-1]
            deltas = attrs["scaleDelta"][::-1]
            coord_names = attrs["unitNames"][::-1]
            coord_lengths = dataset[data_name].shape[:n_dims]

            all_coords_for_data = []
            for dim in range(n_dims):
                coord_name = coord_names[dim]
                delta = (
                    round(deltas[dim]) if coord_names[dim] == "pixel" else deltas[dim]
                )
                coord_values = np.linspace(
                    offsets[dim],
                    offsets[dim] + delta * coord_lengths[dim],
                    coord_lengths[dim],
                    endpoint=False,
                )

                if coord_name in constructed_coords:
                    if np.array_equal(coord_values, constructed_coords[coord_name]):
                        all_coords_for_data.append(coord_name)
                        continue
                    else:
                        coord_name = f"{coord_name}_{data_name}"

                all_coords_for_data.append(coord_name)
                constructed_coords[coord_name] = coord_values
            all_coords_for_data += list(scan_coords.keys())
            data_dimensions[data_name] = tuple(all_coords_for_data)
    return constructed_coords, data_dimensions, scan_coord_names


def get_attrs(hdf5: h5py.File) -> dict[str, str]:
    """Gets the relevant attributes from the HDF5 file"""

    attrs = {}
    try:
        comments = hdf5["Comments"]["PreScan"]
        attrs["comment"] = comments[0][0].decode("ascii")
    except KeyError:
        pass

    def clean_attr(value: bytes) -> float | str:
        value = value.decode("ascii")
        try:
            return float(value)
        except ValueError:
            return value.strip("'")

    column_for_name = {"Beamline": 3}
    skip_headers = ["Low_Level_Scan", "Scan", "Switch"]
    headers = hdf5["Headers"]
    for header in headers:
        if header in skip_headers:
            continue

        header_attrs = {
            item[column_for_name.get(header, 0)].decode("ascii"): clean_attr(item[2])
            for item in headers[header]
        }
        attrs.update(header_attrs)

    return attrs


def dataset_to_array(dataset: h5py.Dataset, type: str = "float64") -> np.ndarray:
    """Quickly converts a h5py dataset to a numpy array"""
    # Converting to a numpy array without using .astype takes ~1000x longer
    return dataset.astype(type)[:]



"""Defines common region selections which are used elsewhere programmatically."""

from enum import Enum


__all__ = ["REGIONS", "normalize_region", "DesignatedRegions"]


class DesignatedRegions(Enum):
    """Commonly used regions which can be used to select data programmatically."""

    # angular windows
    NARROW_ANGLE = 0  # trim to a narrow central region in the spectrometer
    WIDE_ANGLE = 1  # trim to just inside edges of spectrometer data
    TRIM_EMPTY = 2  # trim to edges of spectrometer data

    # energy windows
    BELOW_EF = 10  # everything below e_F
    ABOVE_EF = 11  # everything above e_F
    EF_NARROW = 12  # narrow cut around e_F
    MESO_EF = 13  # comfortably below e_F, pun on mesosphere

    # effective energy windows, determined by Canny edge detection
    BELOW_EFFECTIVE_EF = 20  # everything below e_F
    ABOVE_EFFECTIVE_EF = 21  # everything above e_F
    EFFECTIVE_EF_NARROW = 22  # narrow cut around e_F
    MESO_EFFECTIVE_EF = 23  # comfortably below effective e_F, pun on mesosphere


REGIONS = {
    "copper_prior": {
        "eV": DesignatedRegions.MESO_EFFECTIVE_EF,
    },
    # angular can refer to either 'pixels' or 'phi'
    "wide_angular": {
        # angular can refer to either 'pixels' or 'phi'
        "angular": DesignatedRegions.WIDE_ANGLE,
    },
    "narrow_angular": {
        "angular": DesignatedRegions.NARROW_ANGLE,
    },
}


def normalize_region(region: str | dict):
    """Converts named regions to an actual region."""
    if isinstance(region, str):
        return REGIONS[region]

    if isinstance(region, dict):
        return region

    raise TypeError(
        "Region should be either a string (i.e. an ID/alias) "
        "or an explicit dictionary."
    )



"""Helper functions for coordinate transformations and user/analysis API.

All the functions here assume standard polar angles, as given in the
`data model documentation <https://arpes.readthedocs.io/spectra>`_.

Functions here must accept constants or numpy arrays as valid inputs,
so all standard math functions have been replaced by their equivalents out
of numpy. Array broadcasting should handle any issues or weirdnesses that
would encourage the use of direct iteration, but in case you need to write
a conversion directly, be aware that any functions here must work on arrays
as well for consistency with client code.

Everywhere:

Kinetic energy -> 'kinetic_energy'
Binding energy -> 'eV', for convenience (negative below 0)
Photon energy -> 'hv'

Better facilities should be added for ToFs to do simultaneous (timing, angle) to (binding energy, k-space).
"""

from arpes.utilities.conversion.grids import (
    determine_axis_type,
    determine_momentum_axes_from_measurement_axes,
    is_dimension_unconvertible,
)
from .fast_interp import Interpolator

from typing import Iterable
import collections
import warnings

import numpy as np
import scipy.interpolate

import xarray as xr
from typing import Optional

from .kx_ky_conversion import ConvertKxKy, ConvertKp
from .kz_conversion import ConvertKpKz

__all__ = ["convert_to_kspace", "slice_along_path"]


def grid_interpolator_from_dataarray(
    arr: xr.DataArray,
    fill_value=0.0,
    method="linear",
    bounds_error=False,
):
    """Translates an xarray.DataArray contents into a scipy.interpolate.RegularGridInterpolator.

    This is principally used for coordinate translations.
    """
    flip_axes = set()
    for d in arr.dims:
        c = arr.coords[d]
        if len(c) > 1 and c[1] - c[0] < 0:
            flip_axes.add(d)

    values = arr.values
    for dim in flip_axes:
        values = np.flip(values, arr.dims.index(dim))

    interp_points = [
        arr.coords[d].values[::-1] if d in flip_axes else arr.coords[d].values
        for d in arr.dims
    ]

    if method == "linear":
        return Interpolator.from_arrays(interp_points, values)

    return scipy.interpolate.RegularGridInterpolator(
        points=interp_points,
        values=values,
        bounds_error=bounds_error,
        fill_value=fill_value,
        method=method,
    )


def slice_along_path(
    arr: xr.DataArray,
    interpolation_points=None,
    axis_name=None,
    resolution=None,
    shift_gamma=True,
    n_points: Optional[int] = None,
    extend_to_edge=False,
    **kwargs,
):
    """Gets a cut along a path specified by waypoints in an array.

    TODO: There might be a little bug here where the last coordinate has a value of 0, causing the interpolation to loop
    back to the start point. For now I will just deal with this in client code where I see it until I understand if it is
    universal.

    Interpolates along a path through a volume. If the volume is higher dimensional than the desired path, the
    interpolation is broadcasted along the free dimensions. This allows one to specify a k-space path and receive
    the band structure along this path in k-space.

    Points can either by specified by coordinates, or by reference to symmetry points, should they exist in the source
    array. These symmetry points are translated to regular coordinates immediately, but are provided as a convenience.
    If not all points specify the same set of coordinates, an attempt will be made to unify the coordinates. As an example,
    if the specified path is (kx=0, ky=0, T=20) -> (kx=1, ky=1), the path will be made between (kx=0, ky=0, T=20) ->
    (kx=1, ky=1, T=20). On the other hand, the path (kx=0, ky=0, T=20) -> (kx=1, ky=1, T=40) -> (kx=0, ky=1) will result
    in an error because there is no way to break the ambiguity on the temperature for the last coordinate.

    A reasonable value will be chosen for the resolution, near the maximum resolution of any of the interpolated
    axes by default.

    This function transparently handles the entire path. An alternate approach would be to convert each segment
    separately and concatenate the interpolated axis with xarray.

    If the sentinel value 'G' for the Gamma point is included in the interpolation points, the coordinate axis of the
    interpolated coordinate will be shifted so that its value at the Gamma point is 0. You can opt out of this with the
    parameter 'shift_gamma'

    Args:
        arr: Source data
        interpolation_points: Path vertices
        axis_name: Label for the interpolated axis. Under special
            circumstances a reasonable name will be chosen,
        resolution: Requested resolution along the interpolated axis.
        shift_gamma: Controls whether the interpolated axis is shifted
            to a value of 0 at Gamma.
        n_points: The number of desired points along the output path. This will be inferred
            approximately based on resolution if not provided.
        extend_to_edge: Controls whether or not to scale the vector S -
            G for symmetry point S so that you interpolate
        **kwargs
    such as when the interpolation dimensions are kx and ky: in this case the interpolated dimension will be labeled kp.
    In mixed or ambiguous situations the axis will be labeled by the default value 'inter'.
    to the edge of the available data

    Returns:
        xr.DataArray containing the interpolated data.
    """
    if interpolation_points is None:
        raise ValueError("You must provide points specifying an interpolation path")

    def extract_symmetry_point(name):
        raw_point = arr.attrs["symmetry_points"][name]
        G = arr.attrs["symmetry_points"]["G"]

        if not extend_to_edge or name == "G":
            return raw_point

        # scale the point so that it reaches the edge of the dataset
        S = np.array([raw_point[d] for d in arr.dims if d in raw_point])
        G = np.array([G[d] for d in arr.dims if d in raw_point])

        scale_factor = np.inf
        for i, d in enumerate([d for d in arr.dims if d in raw_point]):
            dS = (S - G)[i]
            coord = arr.coords[d]

            if np.abs(dS) < 0.001:
                continue

            if dS < 0:
                required_scale = (np.min(coord) - G[i]) / dS
                if required_scale < scale_factor:
                    scale_factor = float(required_scale)
            else:
                required_scale = (np.max(coord) - G[i]) / dS
                if required_scale < scale_factor:
                    scale_factor = float(required_scale)

        S = (S - G) * scale_factor + G
        return dict(zip([d for d in arr.dims if d in raw_point], S))

    parsed_interpolation_points = [
        (
            x
            if isinstance(x, Iterable) and not isinstance(x, str)
            else extract_symmetry_point(x)
        )
        for x in interpolation_points
    ]

    free_coordinates = list(arr.dims)
    seen_coordinates = collections.defaultdict(set)
    for point in parsed_interpolation_points:
        for coord, value in point.items():
            seen_coordinates[coord].add(value)
            if coord in free_coordinates:
                free_coordinates.remove(coord)

    for point in parsed_interpolation_points:
        for coord, values in seen_coordinates.items():
            if coord not in point:
                if len(values) != 1:
                    raise ValueError(
                        "Ambiguous interpolation waypoint broadcast at dimension {}".format(
                            coord
                        )
                    )
                else:
                    point[coord] = list(values)[0]

    if axis_name is None:
        try:
            axis_name = determine_axis_type(seen_coordinates.keys())
        except KeyError:
            axis_name = "inter"

        if axis_name == "angle" or axis_name == "inter":
            warnings.warn(
                "Interpolating along axes with different dimensions "
                "will not include Jacobian correction factor."
            )

    converted_coordinates = None
    converted_dims = free_coordinates + [axis_name]

    path_segments = list(
        zip(parsed_interpolation_points, parsed_interpolation_points[1:])
    )

    def element_distance(waypoint_a, waypoint_b):
        delta = np.array([waypoint_a[k] - waypoint_b[k] for k in waypoint_a.keys()])
        return np.linalg.norm(delta)

    def required_sampling_density(waypoint_a, waypoint_b):
        ks = waypoint_a.keys()
        dist = element_distance(waypoint_a, waypoint_b)
        delta = np.array([waypoint_a[k] - waypoint_b[k] for k in ks])
        delta_idx = [
            abs(d / (arr.coords[k][1] - arr.coords[k][0])) for d, k in zip(delta, ks)
        ]
        return dist / np.max(delta_idx)

    # Approximate how many points we should use
    segment_lengths = [element_distance(*segment) for segment in path_segments]
    path_length = sum(segment_lengths)

    gamma_offset = 0  # offset the gamma point to a k coordinate of 0 if possible
    if "G" in interpolation_points and shift_gamma:
        gamma_offset = sum(segment_lengths[0 : interpolation_points.index("G")])

    if resolution is None:
        if n_points is None:
            resolution = np.min(
                [required_sampling_density(*segment) for segment in path_segments]
            )
        else:
            path_length / n_points

    def converter_for_coordinate_name(name):
        def raw_interpolator(*coordinates):
            return coordinates[free_coordinates.index(name)]

        if name in free_coordinates:
            return raw_interpolator

        # Conversion involves the interpolated coordinates
        def interpolated_coordinate_to_raw(*coordinates):
            # Coordinate order is [*free_coordinates, interpolated]
            interpolated = coordinates[len(free_coordinates)] + gamma_offset

            # Start with empty array that we will mask writes onto
            # We need to go with a masking approach rather than a concatenation based one because the coordinates
            # come from np.meshgrid
            dest_coordinate = np.zeros(shape=interpolated.shape)

            start = 0
            for i, l in enumerate(segment_lengths):
                end = start + l
                normalized = (interpolated - start) / l
                seg_start, seg_end = path_segments[i]
                dim_start, dim_end = seg_start[name], seg_end[name]
                mask = np.logical_and(normalized >= 0, normalized < 1)
                dest_coordinate[mask] = (
                    dim_start * (1 - normalized[mask]) + dim_end * normalized[mask]
                )
                start = end

            return dest_coordinate

        return interpolated_coordinate_to_raw

    converted_coordinates = {d: arr.coords[d].values for d in free_coordinates}

    if n_points is None:
        n_points = int(path_length / resolution)

    # Adjust this coordinate under special circumstances
    converted_coordinates[axis_name] = (
        np.linspace(0, path_length, int(path_length / resolution)) - gamma_offset
    )

    converted_ds = convert_coordinates(
        arr,
        converted_coordinates,
        {
            "dims": converted_dims,
            "transforms": dict(
                zip(arr.dims, [converter_for_coordinate_name(d) for d in arr.dims])
            ),
        },
        as_dataset=True,
    )

    if axis_name in arr.dims and len(parsed_interpolation_points) == 2:
        if (
            parsed_interpolation_points[1][axis_name]
            < parsed_interpolation_points[0][axis_name]
        ):
            # swap the sign on this axis as a convenience to the caller
            converted_ds.coords[axis_name].data = -converted_ds.coords[axis_name].data

    if "id" in converted_ds.attrs:
        del converted_ds.attrs["id"]

    return converted_ds


def convert_to_kspace(
    dataset: xr.Dataset,
    bounds=None,
    resolution=None,
    calibration=None,
    coords=None,
    allow_chunks: bool = False,
    **kwargs,
) -> xr.Dataset:
    """Converts volumetric the data to momentum space ("backwards"). Typically what you want.

    Works in general by regridding the data into the new coordinate space and then
    interpolating back into the original data.

    For forward conversion, see sibling methods. Forward conversion works by
    converting the coordinates, rather than by interpolating
    the data. As a result, the data will be totally unchanged by the conversion
    (if we do not apply a Jacobian correction), but the coordinates will no
    longer have equal spacing.

    This is only really useful for zero and one dimensional data because for two dimensional data,
    the coordinates must become two dimensional in order to fully specify every data point
    (this is true in generality, in 3D the coordinates must become 3D as well).

    The only exception to this is if the extra axes do not need to be k-space converted. As is the
    case where one of the dimensions is `cycle` or `delay`, for instance.

    You can request a particular resolution for the new data with the `resolution=` parameter,
    or a specific set of bounds with the `bounds=`

    Examples:
        Convert a 2D cut with automatically inferred range and resolution.

        >>> convert_to_kspace(arpes.io.load_example_data())  # doctest: +SKIP
        xr.DataArray(...)

        Convert a 3D map with a specified momentum window

        >>> convert_to_kspace(  # doctest: +SKIP
                fermi_surface_map,
                kx=np.linspace(-1, 1, 200),
                ky=np.linspace(-1, 1, 350),
            )
        xr.DataArray(...)

    Args:
        arr (xr.DataArray): [description]
        #bounds ([type], optional): [description]. Defaults to None.
        resolution ([type], optional): [description]. Defaults to None.
        calibration ([type], optional): [description]. Defaults to None.
        coords ([type], optional): [description]. Defaults to None.
        allow_chunks (bool, optional): [description]. Defaults to False.

    Raises:
        NotImplementedError: [description]
        AnalysisError: [description]
        ValueError: [description]

    Returns:
        [type]: [description]
    """
    if coords is None:
        coords = {}
    coords.update(kwargs)

    spectrum: xr.DataArray = dataset.S.spectrum

    has_eV = "eV" in dataset.dims

    # Chunking logic
    if allow_chunks and has_eV and len(dataset.eV) > 50:
        DESIRED_CHUNK_SIZE = 1000 * 1000 * 20
        n_chunks = np.prod(spectrum.shape) // DESIRED_CHUNK_SIZE
        if n_chunks > 100:
            warnings.warn("Input array is very large. Please consider resampling.")

        chunk_thickness = max(len(dataset.eV) // n_chunks, 1)

        finished = []
        low_idx = 0
        high_idx = chunk_thickness

        while low_idx < len(dataset.eV):
            chunk = dataset.isel(eV=slice(low_idx, high_idx))

            if len(chunk.eV) == 1:
                chunk = chunk.squeeze("eV")

            kchunk = convert_to_kspace(
                chunk,
                bounds=bounds,
                resolution=resolution,
                calibration=calibration,
                coords=coords,
                allow_chunks=False,
                **kwargs,
            )

            if "eV" not in kchunk.dims:
                kchunk = kchunk.expand_dims("eV")

            finished.append(kchunk)

            low_idx = high_idx
            high_idx = min(len(dataset.eV), high_idx + chunk_thickness)

        return xr.concat(finished, dim="eV")

    # Chunking is finished here

    # TODO be smarter about the resolution inference
    removed = [d for d in spectrum.dims if is_dimension_unconvertible(d)]
    old_dims = [d for d in spectrum.dims if not is_dimension_unconvertible(d)]
    old_dims.sort()

    # Energy gets put at the front as a standardization
    if "eV" in removed:
        removed.remove("eV")

    # temporarily reassign coordinates for dimensions we will not
    # convert to "index-like" dimensions
    restore_index_like_coordinates = {r: dataset.coords[r].values for r in removed}
    new_index_like_coordinates = {
        r: np.arange(len(dataset.coords[r].values)) for r in removed
    }
    dataset = dataset.assign_coords(**new_index_like_coordinates)

    if not old_dims:
        return dataset  # no need to convert, might be XPS or similar

    converted_dims = (
        (["eV"] if has_eV else [])
        + determine_momentum_axes_from_measurement_axes(old_dims)
        + removed
    )

    convert_cls = {
        ("phi",): ConvertKp,
        ("beta", "phi"): ConvertKxKy,
        ("phi", "theta"): ConvertKxKy,
        ("phi", "psi"): ConvertKxKy,
        # ('chi', 'phi',): ConvertKxKy,
        ("hv", "phi"): ConvertKpKz,
    }.get(tuple(old_dims))
    converter: ConvertKp | ConvertKxKy | ConvertKpKz = convert_cls(
        dataset, converted_dims, calibration=calibration
    )

    converted_coordinates = converter.get_coordinates(
        resolution=resolution, bounds=bounds
    )

    if not set(coords.keys()).issubset(converted_coordinates.keys()):
        extra = set(coords.keys()).difference(converted_coordinates.keys())
        raise ValueError("Unexpected passed coordinates: {}".format(extra))

    converted_coordinates.update(coords)

    result = convert_coordinates(
        dataset,
        converted_coordinates,
        {
            "dims": converted_dims,
            "transforms": dict(
                zip(dataset.dims, [converter.conversion_for(d) for d in dataset.dims])
            ),
        },
    )
    result = result.assign_coords(**restore_index_like_coordinates)
    return result


def convert_coordinates(
    ds: xr.Dataset,
    target_coordinates: dict,
    coordinate_transform: dict,
) -> xr.Dataset:
    ds = ds.copy(deep=True)
    ordered_source_dimensions = ds.dims
    grid_interpolator = grid_interpolator_from_dataarray(
        ds.S.spectrum.transpose(*ordered_source_dimensions),
        fill_value=float("nan"),
    )

    # Skip the Jacobian correction for now
    # Convert the raw coordinate axes to a set of gridded points
    meshed_coordinates = np.meshgrid(
        *[target_coordinates[dim] for dim in coordinate_transform["dims"]],
        indexing="ij",
    )
    meshed_coordinates = [meshed_coord.ravel() for meshed_coord in meshed_coordinates]

    if "eV" not in ds.dims:
        try:
            meshed_coordinates = [ds.S.lookup_offset_coord("eV")] + meshed_coordinates
        except ValueError:
            pass

    ordered_transformations = [
        coordinate_transform["transforms"][dim] for dim in ds.dims
    ]

    transformed_coordinates = []
    for tr in ordered_transformations:
        transformed_coordinates.append(tr(*meshed_coordinates))

    if not isinstance(grid_interpolator, Interpolator):
        transformed_coordinates = np.array(transformed_coordinates).T

    converted_volume = grid_interpolator(transformed_coordinates)

    # Wrap it all up
    def acceptable_coordinate(c: np.ndarray | xr.DataArray) -> bool:
        # Currently we do this to filter out coordinates that are functions of the old angular dimensions,
        # we could forward convert these, but right now we do not
        try:
            if set(c.dims).issubset(coordinate_transform["dims"]):
                return True
            else:
                return False
        except:
            return True

    target_coordinates = {
        k: v for k, v in target_coordinates.items() if acceptable_coordinate(v)
    }

    converted_spectrum = np.reshape(
        converted_volume,
        [len(target_coordinates[d]) for d in coordinate_transform["dims"]],
        order="C",
    )
    spectrum_name = ds.S.spectrum.name
    spectrum = xr.DataArray(
        converted_spectrum, target_coordinates, coordinate_transform["dims"]
    )
    ds = ds.drop_vars(spectrum_name)
    ds = ds.assign({spectrum_name: spectrum})

    return ds



"""Contains specifications and paths to BZs for some common materials.

These are
useful if you know them ahead of hand, either from a picture, from a DFT
calculation, tight binding calculation, or explicit specification.

This is used in the interactive BZ explorer in order to help with orienting
yourself in momentum.

Each zone definition corresponds to the following things:

1. The geometry, this gives the physical Brillouin zone size and shape
2. The material work function, if available
3. The material inner potential, if available
4. The material name
"""

import functools
import numpy as np
import pathlib

from arpes.utilities.bz import hex_cell_2d, as_2d

A_GRAPHENE = 2.46 / (2 * np.pi)
A_WS2 = 3.15 / (2 * np.pi)
A_WSe2 = 3.297 / (2 * np.pi)


def bz_points_for_hexagonal_lattice(a=1):
    """Calculates the Brillouin zone corners for a triangular (colloq. hexagona) lattice."""
    from ase.dft.bz import bz_vertices

    cell = hex_cell_2d(a)
    cell = [list(c) + [0] for c in cell] + [[0, 0, 1]]
    icell = np.linalg.inv(cell).T
    bz_vertices = bz_vertices(icell)

    # get the first face which has six points, this is the top or bottom
    # face of the cell
    return as_2d(bz_vertices[[len(face[0]) for face in bz_vertices].index(6)][0])


def image_for(file):
    """Loads a preview image showing the Brillouin zone shape."""
    f = pathlib.Path(__file__).parent / ".." / "example_data" / "brillouin_zones" / file
    return str(f.absolute())


SURFACE_ZONE_DEFINITIONS = {  # : Dict[str, Dict[str, any]]
    "2H-WS2": {
        "name": "2H-Tungsten Disulfide",
        "work_function": None,
        "inner_potential": None,
        "bz_points": functools.partial(bz_points_for_hexagonal_lattice, a=A_WS2),
    },
    "Graphene": {
        "name": "Graphene",
        "work_function": None,
        "inner_potential": None,
        "bz_points": functools.partial(bz_points_for_hexagonal_lattice, a=A_GRAPHENE),
    },
    "2H-WSe2": {
        "name": "Tungsten Diselenide",
        "work_function": None,
        "inner_potential": None,
        "bz_points": functools.partial(bz_points_for_hexagonal_lattice, a=A_WS2),
    },
    "1T-TiSe2": {
        "name": "1T-Titanium Diselenide",
        "work_function": None,
        "inner_potential": None,
        "image": image_for("1t-tise2-bz.png"),
        "image_waypoints": [
            # everywhere waypoints are pixel_x, pixel_y, mom_x, mom_y
            # two waypoints are requried in order to specify
            [],
            [],
        ],
        "image_src": "https://arxiv.org/abs/1712.04967",
    },
    "Td-WTe2": {
        "name": "Td-Tungsten Ditelluride",
        "work_function": None,
        "inner_potential": None,
        "image": image_for("td-wte2-bz.png"),
        "image_waypoints": [
            [445, 650, -0.4, -0.2],
            [1470, 166, 0.4, 0.2],
        ],
        "image_src": "https://arxiv.org/abs/1603.08508",
    },
    "NCCO": {
        "name": "Nd_{2-x}Ce_xCuO_4",
        "work_function": None,
        "inner_potential": None,
        "image": image_for("cuprate-bz.png"),
        "image_waypoints": [
            [],
            [],
        ],
        "image_src": "https://vishiklab.faculty.ucdavis.edu/wp-content/uploads/sites/394/2016/12/ARPES-studies-of-cuprates-online.pdf",
    },
    "Bi2212": {
        "name": "Bi_2Sr_2CaCu_2O_{8+x}",
        "work_function": None,
        "inner_potential": None,
        "image": image_for("cuprate-bz.png"),
        "image_waypoints": [
            [],
            [],
        ],
        "image_src": "https://vishiklab.faculty.ucdavis.edu/wp-content/uploads/sites/394/2016/12/ARPES-studies-of-cuprates-online.pdf",
    },
    "1H-NbSe2": {
        "name": "1H-Niobium Diselenide",
        "work_function": None,
        "inner_potential": None,
        "image": image_for("1h-nbse2-bz.png"),
        "image_waypoints": [
            [],
            [],
        ],
        "image_src": "https://www.nature.com/articles/s41467-018-03888-4",
    },
    "1H-TaS2": {
        "name": "1H-Tantalum Disulfide",
        "work_function": None,
        "inner_potential": None,
        "image": image_for("1h-tas2-bz.png"),
        "image_waypoints": [
            [],
            [],
        ],
        "image_src": "https://www.nature.com/articles/s41467-018-03888-4",
    },
}



"""Implements data loading for the Lanzara group "Main Chamber"."""

import numpy as np
import xarray as xr

from arpes.endstations import FITSEndstation, HemisphericalEndstation
from arpes.constants import pi

__all__ = ("ALGMainChamber",)


class ALGMainChamber(HemisphericalEndstation, FITSEndstation):
    """Implements data loading for the Lanzara group "Main Chamber"."""

    PRINCIPAL_NAME = "ALG-Main"
    ALIASES = [
        "MC",
        "ALG-Main",
        "ALG-MC",
        "ALG-Hemisphere",
        "ALG-Main Chamber",
    ]
    ATTR_TRANSFORMS = {
        "START_T": lambda l: {
            "time": " ".join(l.split(" ")[1:]).lower(),
            "date": l.split(" ")[0],
        },
    }

    RENAME_KEYS = {
        "Phi": "chi",
        "Beta": "beta",
        "Theta": "theta",
        "Azimuth": "chi",
        "Alpha": "alpha",
        "Pump_energy_uJcm2": "pump_fluence",
        "T0_ps": "t0_nominal",
        "W_func": "workfunction",
        "Slit": "slit",
        "LMOTOR0": "x",
        "LMOTOR1": "y",
        "LMOTOR2": "z",
        "LMOTOR3": "theta",
        "LMOTOR4": "beta",
        "LMOTOR5": "chi",
        "LMOTOR6": "delay",
        "SFLNM0": "lens_mode_name",
        "SFFR_0": "frames_per_slice",
        "SFBA_0": "phi_prebinning",
        "SFBE0": "eV_prebinning",
        "null": "cycle",
        "X": "x",
        "Y": "y",
    }

    MERGE_ATTRS = {
        "analyzer": "Specs PHOIBOS 150",
        "analyzer_name": "Specs PHOIBOS 150",
        "parallel_deflectors": False,
        "perpendicular_deflectors": False,
        "analyzer_radius": 150,
        "analyzer_type": "hemispherical",
        "mcp_voltage": None,
        "probe_linewidth": 0.015,
    }

    RAD_PER_PIXEL = (1 / 10) * (pi / 180)

    def postprocess_frame(self, frame: xr.Dataset):
        frame = super().postprocess_frame(frame)

        if "pixel" in frame.coords:
            phi_axis = frame.coords["pixel"].values * self.RAD_PER_PIXEL

            if "pixel" in frame.coords:
                frame = frame.rename(pixel="phi")

            frame = frame.assign_coords(phi=phi_axis)
        return frame

    def postprocess_scan(self, data: xr.Dataset, scan_desc: dict = None):
        """Performs final normalization of scan data.

        For the Lanzaa group main chamber, this means:

        1. Associating the fixex UV laser energy.
        2. Adding missing coordinates.
        3. Using a standard approximate set of coordinate offsets.
        4. Converting relevant angular coordinates to radians.
        """
        data.attrs["hv"] = 5.93
        data.attrs["alpha"] = 0
        data.attrs["psi"] = 0
        # by default we use this value since this isnear the center of the spectrometer window
        data.attrs["phi_offset"] = 0.405

        data = super().postprocess_scan(data, scan_desc)

        if "beta" in data.coords:
            data = data.assign_coords(beta=data.beta.values * np.pi / 180)

        return data



"""This module contains methods that get unitful alignments of one array against another.

This is very useful for determining spectral shifts before doing serious curve fitting analysis or similar.

Implementations are included for each of 1D and 2D arrays, but this could be simply extended to ND if we need to.
I doubt that this is necessary and don't mind the copied code too much at the present.
"""

import numpy as np
import xarray as xr

from scipy import signal
from arpes.fits.fit_models import QuadraticModel

__all__ = ("align2d", "align1d", "align")


def align2d(a, b, subpixel=True):
    """Returns the unitful offset of b in a for 2D arrays using 2D correlation.

    Args:
        a: The first input array.
        b: The second input array.
        subpixel: If True, will perform subpixel alignment using a curve fit to the peak.

    Returns:
        The offset of a 2D array against another.
    """
    corr = signal.correlate2d(
        a.values - np.mean(a.values), b.values - np.mean(b.values), boundary="fill", mode="same"
    )

    y, x = np.unravel_index(np.argmax(corr), corr.shape)

    if subpixel:
        marg = xr.DataArray(
            corr[y - 10 : y + 10, x], coords={"index": np.linspace(-10, 9, 20)}, dims=["index"]
        )
        marg = marg / np.max(marg)
        mod = QuadraticModel().guess_fit(marg)
        true_y = y + -mod.params["b"].value / (2 * mod.params["a"].value)

        marg = xr.DataArray(
            corr[y, x - 10 : x + 10], coords={"index": np.linspace(-10, 9, 20)}, dims=["index"]
        )
        marg = marg / np.max(marg)
        mod = QuadraticModel().guess_fit(marg)
        true_x = x + -mod.params["b"].value / (2 * mod.params["a"].value)

        y, x = true_y, true_x

    y = 1.0 * y - a.values.shape[0] / 2.0
    x = 1.0 * x - a.values.shape[1] / 2.0

    return (
        y * a.G.stride(generic_dim_names=False)[a.dims[0]],
        x * a.G.stride(generic_dim_names=False)[a.dims[1]],
    )


def align1d(a, b, subpixel=True):
    """Returns the unitful offset of b in a for 1D arrays using 1D correlation.

    Args:
        a: The first input array.
        b: The second input array.
        subpixel: If True, will perform subpixel alignment using a curve fit to the peak.

    Returns:
        The offset of an array against another.
    """
    corr = np.correlate(a.values - np.mean(a.values), b.values - np.mean(b.values), mode="same")
    (x,) = np.unravel_index(np.argmax(corr), corr.shape)

    if subpixel:
        marg = xr.DataArray(
            corr[x - 10 : x + 10], coords={"index": np.linspace(-10, 9, 20)}, dims=["index"]
        )
        marg = marg / np.max(marg)
        mod = QuadraticModel().guess_fit(marg)
        x = x + -mod.params["b"].value / (2 * mod.params["a"].value)

    x = 1.0 * x - a.values.shape[0] / 2.0
    return x * a.G.stride(generic_dim_names=False)[a.dims[0]]


def align(a, b, **kwargs):
    """Returns the unitful offset of b in a for ndarrays.

    Args:
        a: The first input array.
        b: The second input array.
        subpixel: If True, will perform subpixel alignment using a curve fit to the peak.

    Returns:
        The offset of an array against another.
    """
    if len(a.dims == 1):
        return align1d(a, b, **kwargs)

    assert len(a.dims) == 2
    return align2d(a, b, **kwargs)



"""An axis binning control."""

from PyQt5 import QtWidgets

from arpes.utilities.qt.ui import layout

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from weakref import ReferenceType
    from .base import ImageTool

__all__ = ("BinningInfoWidget",)


class BinningInfoWidget(QtWidgets.QGroupBox):
    """A spinbox allowing you to set the binning on different axes."""

    def __init__(
        self, parent=None, root: "ReferenceType" = None, axis_name: str = None
    ):
        """Initialize an inner spinbox and connect signals to get reactivity."""
        super().__init__(title=axis_name, parent=parent)
        self._root = root
        self.axis_name = axis_name

        self.spinbox = QtWidgets.QSpinBox()
        self.spinbox.setMaximum(2000)
        self.spinbox.setMinimum(1)
        self.spinbox.setValue(1)
        self.spinbox.valueChanged.connect(self.changeBinning)
        self.spinbox.editingFinished.connect(self.changeBinning)

        self.layout = layout(
            self.spinbox,
            widget=self,
        )

        self.recompute()

    @property
    def root(self) -> "ImageTool":
        """Unwraps the weakref to the parent application."""
        return self._root()

    def recompute(self):
        """Redraws all dependent UI state, namely the title."""
        self.setTitle(self.axis_name)

    def changeBinning(self):
        """Callback for widget value changes which sets the binning on the root app."""
        try:
            old_binning = self.root.binning.copy()
            old_binning[self.axis_name] = self.spinbox.value()
            self.root.binning = old_binning
        except:
            pass
